package com.capitalware.ufm.action.handlers;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;

import org.jdom.JDOMException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.*;
import com.capitalware.ufm.actions.*;
import com.capitalware.ufm.handlers.InternalMessageQueue;
import com.capitalware.ufm.internal.messages.*;
import com.capitalware.ufm.mq.MQStatusUpdate;
import com.capitalware.ufm.mq.objects.MQObject;
import com.capitalware.ufm.mq.objects.StatusObject;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.watch.Watch;
import com.capitalware.ufm.xml.MQXml;

/**
 * This class will process a list of action commands.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ProcessActions
{
   private static final String LINE_SEPARATOR = System.getProperty("line.separator");
   
   public  Hashtable<String,String> actionProperties = null;
   private Watch                    watchClass = null;
   private MQStatusUpdate           mqsu = null;
   private InternalMessageQueue     iq = null;
   private boolean                  internalActions = false;
   
   /**
    * The constructor
    * @param iq
    */
   public ProcessActions(InternalMessageQueue iq)
   {
      super();
      this.iq = iq;
      internalActions = true;
   }
   
   /**
    * The constructor
    * @param statusObject
    */
   public ProcessActions(StatusObject statusObject)
   {
      super();
      
      if (statusObject != null)
         iq = handleStatusQueue(statusObject);
   }
   
   /**
    * Get the Internal Queue
    * @return iq
    */
   public InternalMessageQueue getInternalQueue()
   {
      return iq;
   }

   /**
    * Loop through a list of action commands and process each one.
    * @param actionAL
    * @param actionProperties
    * @return exitCode
    */
   public int doWork(ArrayList<UFMObject> actionAL, Hashtable<String,String> actionProperties)
   {
      this.actionProperties = actionProperties;
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      int      exitCode = 0;
      UFMObject fmo = null;

      /**
       * First check to make sure all of the objects are ok.
       */
      try
      {
         for (int i=0; i < actionAL.size(); i++)
         {
            fmo = actionAL.get(i);

            /* Give the action the ability to do a callback */
            fmo.paClass = this;
            
            fmo.performTokenization();
            fmo.checkObject();
         }
      }
      catch (IllegalArgumentException e)
      {
         UFM.logger.fatal("IllegalArgumentException: " + fmo.actionName + " Action : " + e.getLocalizedMessage());
         UFM.globalErrorText = fmo.actionName + " Action : " + e.getLocalizedMessage();
         exitCode = 16;
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + fmo.actionName + " Action : " + e.getLocalizedMessage());
         UFM.globalErrorText = fmo.actionName + " Action : " + e.getLocalizedMessage();
         exitCode = 16;
      }

      /**
       * If there was an issue in the checkObject method, do not continue
       */
      if (exitCode != 0)
         return exitCode;

      /**
       * Life's good - now process each action command.
       */
      try
      {
         for (int i=0; (i < actionAL.size()) && (exitCode == 0); i++)
         {
            UFM.globalErrorText = null;
            fmo = actionAL.get(i);

            if (fmo instanceof AppendObject)
            {
               AppendAction aa = new AppendAction((AppendObject)fmo);
               exitCode = aa.doWork();
            }
            else if (fmo instanceof ConvertObject)
            {
               ConvertAction ca = new ConvertAction((ConvertObject)fmo);
               exitCode = ca.doWork();
            }
            else if (fmo instanceof CopyObject)
            {
               CopyAction ca = new CopyAction((CopyObject)fmo);
               exitCode = ca.doWork();
            }
            else if (fmo instanceof DecryptFileObject)
            {
               DecryptFileAction dfa = new DecryptFileAction((DecryptFileObject)fmo);
               exitCode = dfa.doWork();
            }
            else if (fmo instanceof DeleteObject)
            {
               DeleteAction da = new DeleteAction((DeleteObject)fmo);
               exitCode = da.doWork();
            }
            else if (fmo instanceof EchoObject)
            {
               EchoAction ea = new EchoAction((EchoObject)fmo);
               exitCode = ea.doWork();
            }
            else if (fmo instanceof EncryptFileObject)
            {
               EncryptFileAction efa = new EncryptFileAction((EncryptFileObject)fmo);
               exitCode = efa.doWork();
            }
            else if (fmo instanceof ExecuteObject)
            {
               ExecuteAction ea = new ExecuteAction((ExecuteObject)fmo, iq);
               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fmo.toStringForStatusUpdate("Starting")));
               exitCode = ea.doWork();
            }
            else if (fmo instanceof FtpObject)
            {
               FtpAction fa = new FtpAction((FtpObject)fmo, iq);
               exitCode = fa.doWork();
            }
            else if (fmo instanceof HttpGetFileObject)
            {
               HttpGetFileAction hgfa = new HttpGetFileAction((HttpGetFileObject)fmo);
               exitCode = hgfa.doWork();
            }
            else if (fmo instanceof HttpPutFileObject)
            {
               HttpPutFileAction hpfa = new HttpPutFileAction((HttpPutFileObject)fmo);
               exitCode = hpfa.doWork();
            }
            else if (fmo instanceof IfObject)
            {
               IfAction ea = new IfAction((IfObject)fmo, iq);
               exitCode = ea.doWork();
            }
            else if (fmo instanceof LaunchObject)
            {
               LaunchAction la = new LaunchAction((LaunchObject)fmo);
               exitCode = la.doWork();
            }
            else if (fmo instanceof LoopObject)
            {
               LoopAction la = new LoopAction((LoopObject)fmo, iq);
               exitCode = la.doWork();
            }
            else if (fmo instanceof MakeDirObject)
            {
               MakeDirAction mda = new MakeDirAction((MakeDirObject)fmo);
               exitCode = mda.doWork();
            }
            else if (fmo instanceof MergeObject)
            {
               MergeAction ma = new MergeAction((MergeObject)fmo);
               exitCode = ma.doWork();
            }
            else if (fmo instanceof MergeSortObject)
            {
               MergeSortAction msa = new MergeSortAction((MergeSortObject)fmo);
               exitCode = msa.doWork();
            }
            else if (fmo instanceof MoveObject)
            {
               MoveAction ma = new MoveAction((MoveObject)fmo);
               exitCode = ma.doWork();
            }
            else if (fmo instanceof MQPutQuitObject)
            {
               MQPutQuitAction mqpqa = new MQPutQuitAction((MQPutQuitObject)fmo);
               exitCode = mqpqa.doWork();
            }
            else if (fmo instanceof MQReceiveObject)
            {
               MQReceiveAction mqra = new MQReceiveAction((MQReceiveObject) fmo, iq);
               exitCode = mqra.doWork();
            }
            else if (fmo instanceof MQSendObject)
            {
               MQSendAction mqsa = new MQSendAction((MQSendObject) fmo, iq);
               exitCode = mqsa.doWork();
            }
            else if (fmo instanceof ReEncodeObject)
            {
               ReEncodeAction rea = new ReEncodeAction((ReEncodeObject)fmo);
               exitCode = rea.doWork();
            }
            else if (fmo instanceof RemoveItemsObject)
            {
               RemoveItemsAction ria = new RemoveItemsAction((RemoveItemsObject)fmo);
               exitCode = ria.doWork();
            }
            else if (fmo instanceof RemoveDirObject)
            {
               RemoveDirAction rda = new RemoveDirAction((RemoveDirObject)fmo);
               exitCode = rda.doWork();
            }
            else if (fmo instanceof RenameObject)
            {
               RenameAction ra = new RenameAction((RenameObject)fmo);
               exitCode = ra.doWork();
            }
            else if (fmo instanceof ReplaceTextObject)
            {
               ReplaceTextAction rta = new ReplaceTextAction((ReplaceTextObject)fmo);
               exitCode = rta.doWork();
            }
            else if (fmo instanceof ScheduleObject)
            {
               ScheduleAction sa = new ScheduleAction((ScheduleObject)fmo, iq);
               exitCode = sa.doWork();
            }
            else if (fmo instanceof ScpObject)
            {
               ScpAction sa = new ScpAction((ScpObject)fmo, iq);
               exitCode = sa.doWork();
            }
            else if (fmo instanceof SendEmailObject)
            {
               SendEmailAction sea = new SendEmailAction((SendEmailObject)fmo, iq);
               exitCode = sea.doWork();
            }
            else if (fmo instanceof SFtpObject)
            {
               SFtpAction fa = new SFtpAction((SFtpObject)fmo, iq);
               exitCode = fa.doWork();
            }
            else if (fmo instanceof SleepObject)
            {
               SleepAction sa = new SleepAction((SleepObject)fmo);
               exitCode = sa.doWork();
            }
            else if (fmo instanceof SortObject)
            {
               SortAction sa = new SortAction((SortObject)fmo);
               exitCode = sa.doWork();
            }
            else if (fmo instanceof SortUniqueObject)
            {
               SortUniqueAction ua = new SortUniqueAction((SortUniqueObject)fmo);
               exitCode = ua.doWork();
            }
            else if (fmo instanceof SplitObject)
            {
               SplitAction ta = new SplitAction((SplitObject)fmo);
               exitCode = ta.doWork();
            }
            else if (fmo instanceof TarObject)
            {
               TarAction ta = new TarAction((TarObject)fmo);
               exitCode = ta.doWork();
            }
            else if (fmo instanceof TouchObject)
            {
               TouchAction ta = new TouchAction((TouchObject)fmo);
               exitCode = ta.doWork();
            }
            else if (fmo instanceof UniqueObject)
            {
               UniqueAction ua = new UniqueAction((UniqueObject)fmo);
               exitCode = ua.doWork();
            }
            else if (fmo instanceof UnTarObject)
            {
               UnTarAction uta = new UnTarAction((UnTarObject)fmo);
               exitCode = uta.doWork();
            }
            else if (fmo instanceof UnZipObject)
            {
               UnZipAction uza = new UnZipAction((UnZipObject)fmo);
               exitCode = uza.doWork();
            }
            else if (fmo instanceof WatchObject)
            {
               WatchAction wa = new WatchAction((WatchObject) fmo, iq);
               if (iq != null)
                  iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+fmo.toStringForStatusUpdate("Starting")));
                  
               exitCode = wa.doWork();
            }
            else if (fmo instanceof WriteTextObject)
            {
               WriteTextAction wta = new WriteTextAction((WriteTextObject)fmo);
               exitCode = wta.doWork();
            }
            else if (fmo instanceof ZipObject)
            {
               ZipAction za = new ZipAction((ZipObject)fmo);
               exitCode = za.doWork();
            }
            else
            {
               UFM.logger.fatal("Unknown action selected: " + fmo.toString());
               exitCode = 99;
            }
            
            UFM.actionObjectValues.put(fmo.labelName+"_"+fmo.actionName, fmo);
            
            if ( (iq != null) && 
                 !(fmo instanceof ExecuteObject) && !(fmo instanceof MQReceiveObject) && !(fmo instanceof MQSendObject)  )
               iq.addMessage(new UpdateStatus(exitCode+"", dateTimeFormat.format(new Date())+", "+exitCode+", "+fmo.toStringForStatusUpdate("??")));
            
            if (exitCode != 0)
            {
               UFM.logger.fatal("Error condition found - UFM will not continue with the current group of Actions.");
               
               UFM.handleOnError(actionProperties, iq);
               UFM.stopAllThreads();
            }
         }
      }
      catch (Exception e)
      {
         UFM.logger.fatal("Exception: " + fmo.actionName + " Action : " + e.getLocalizedMessage());
         exitCode = 16;
         if (iq != null)
            iq.addMessage(new UpdateStatus(exitCode+"", dateTimeFormat.format(new Date())+", "+exitCode+", "+fmo.toStringForStatusUpdate("Error")));
      }

      // If internal actions (i.e. if, loop, watch) then do not send "exiting" message 
      // nor the quit message
      if  (!internalActions)   
      {
         if (iq != null)
            iq.addMessage(new UpdateStatus(exitCode+"", dateTimeFormat.format(new Date())+", "+exitCode+", , ,Exiting, "));
   
         if (iq != null)
            iq.addMessage(new Quit());
         
         
         if (mqsu != null)
         {
            while (mqsu.isRunning())
            {
               try
               {
                 Thread.sleep(250); // 1/4 second
               }
               catch (InterruptedException e) 
               {}
            }
         }
      }
      
      return exitCode;
   }

   /**
    * Handle the connecting to the QMgr and opening the status queue.
    * Also, it will create the share-able internal queue. 
    * @return iq
    */
   private InternalMessageQueue handleStatusQueue(StatusObject statusObject)
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      InternalMessageQueue iq = null;
      String newMsg = "";
      
      try
      {
         if (statusObject.mqXMLFile != null)
         {
            MQXml mqp = new MQXml(statusObject.qMgrName, statusObject.qObject, statusObject.CCDTFile);
            mqp.readXMLFile(statusObject.mqXMLFile);
            statusObject.mqo = mqp.getMQObject();
         }
         else
         {
            statusObject.mqo = new MQObject(statusObject.qMgrName, statusObject.qObject, statusObject.CCDTFile);
         }

         statusObject.mqo.checkQM();
         iq = new InternalMessageQueue(statusObject.mqo.qObject.qName+"_IQ");
         UFM.logger.info("Status Queue Information:");
         statusObject.mqo.dumpUserParms();
         statusObject.checkObject();
         statusObject.printValues();

         mqsu = new MQStatusUpdate(statusObject.mqo);
         mqsu.setTimeToLive(statusObject.timeToLive);
         mqsu.setIQ(iq);
         mqsu.start();
         
         try
         {
            newMsg += InetAddress.getLocalHost().getHostName()+", ";
         }
         catch (UnknownHostException e)
         {
            newMsg += ", ";
         }
         
         try
         {
            newMsg += InetAddress.getLocalHost().getHostAddress()+", ";
         }
         catch (UnknownHostException e)
         {
            newMsg += ", ";
         }

         newMsg += System.getProperty("os.name")+", "+System.getProperty("user.name")+", 0, "+CWConstants.VERSION+", "+UFM.getXmlFilename()+LINE_SEPARATOR;

         iq.addMessage(new NewStatus(dateTimeFormat.format(new Date())+", "+newMsg));
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: " + e.getMessage());
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (JDOMException e)
      {
         UFM.logger.fatal("JDOMException: " + e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         UFM.globalErrorText = e.getLocalizedMessage();
      }
      catch (Exception e)
      {
         UFM.logger.fatal(e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         UFM.globalErrorText = e.getLocalizedMessage();
      }

      return iq;
   }
   
   /**
    * Set the Watch class
    * @param watch
    */
   public void setWatchClass(Watch watch)
   {
      this.watchClass = watch;
   }
   
   /**
    * The "Watch Found" file has been deleted, so handle resetting the file's timestamp
    * on the watch list.
    */
   public void handleDeletedWatchFoundFile()
   {
      if (watchClass != null)
      {
         if (actionProperties.containsKey(CWConstants.WATCH_FOUND_FILE))
         {
            String watchFile = actionProperties.get(CWConstants.WATCH_FOUND_FILE);
            
            UFM.logger.debug("Removing Watch Found File Key " + CWConstants.WATCH_FOUND_FILE + " with value of '"+watchFile+"'");
            /* reset the file's watch timestamp */
            watchClass.resetLastModified(watchFile);
         }
      }
   }
}
