package com.capitalware.ufm.action.objects;

import java.io.File;
import java.util.ArrayList;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.handlers.Tokens;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.objects.Cryptography;

/**
 * This class holds the values for the EncryptFile action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class EncryptFileObject extends UFMFileObject
{
   public String        srcFilename = null;
   public String        toFile = null;
   public String        toDir = null;
   public Cryptography  crypto = new Cryptography();
   public boolean       createDir = false;

   /**
    * The constructor
    */
   public EncryptFileObject()
   {
      super();
   }

   /**
    * Process tokens in Action Object's variables
    */
   public void performTokenization()
   {
      if (srcFilename != null)
         srcFilename   = Tokens.replace(srcFilename, paClass.actionProperties);
      if (toFile != null)
         toFile = Tokens.replace(toFile, paClass.actionProperties);
      if (toDir != null)
         toDir = Tokens.replace(toDir, paClass.actionProperties);
   }

   /**
    * Verify the values in the object.
    */
   public void checkObject()
   {
      checkAndCreateDirectory(createDir, toDir);
      
      if (srcFilename == null)
         throw new IllegalArgumentException("Error: You must specify a filename for the EncryptFile action.");

      if (toFile == null)
         throw new IllegalArgumentException("Error: You must specify a toFilename for the EncryptFile action.");

      if ( (crypto.keySize != 128) && (crypto.keySize != 192) && (crypto.keySize != 256) )
         throw new IllegalArgumentException("Error: You must specify a keysize attribute that is 128 or 192 or 256 for the EncryptFile action.");

      if ("".equals(crypto.passPhrase))
         throw new IllegalArgumentException("Error: PassPhrase attribute cannot be blank for the EncryptFile action.");
   }

   /**
    * Output the info
    * @param logger
    */
   public void printValues()
   {
      UFM.logger.info("     Source File          = " + srcFilename);
      UFM.logger.info("     Target Directory     = " + toDir);
      UFM.logger.info("     Target File          = " + toFile);
      UFM.logger.info("     KeySize              = " + crypto.keySize);
      UFM.logger.info("     PassPhrase           = " + crypto.passPhrase );
      UFM.logger.info("     Create Directory     = " + createDir);
   }
   
   /**
    * Convert the object to a string.
    * @return s
    */
   public ArrayList<String> toStringArray()
   {
      ArrayList<String> al = new ArrayList<String>();
      String file;

      if (srcFilename.indexOf(File.separator) != -1 )
      {
         int x = srcFilename.lastIndexOf(File.separator);

         String dir   = srcFilename.substring(0, x);
         String fname = srcFilename.substring(x+1);

         file = "<"+CWConstants.FILE_ELEMENT+" "+CWConstants.DIR_ATTRIBUTE+"=\""+dir+"\">"+fname+"</"+CWConstants.FILE_ELEMENT+">";
      }
      else
         file = "<"+CWConstants.FILE_ELEMENT+">"+srcFilename+"</"+CWConstants.FILE_ELEMENT+">";
      
      if (toDir == null)
         al.add("<"+CWConstants.ENCRYPT_FILE_ELEMENT+" "+CWConstants.TO_FILE_ATTRIBUTE+"=\""+toFile+" "+CWConstants.KEY_SIZE_ATTRIBUTE+"=\"+crypto.keySize+\" "+CWConstants.PASS_PHRASE_ATTRIBUTE+"=\""+crypto.passPhrase+"\">"+LINE_SEPARATOR);
      else
         al.add("<"+CWConstants.ENCRYPT_FILE_ELEMENT+" "+CWConstants.TO_DIR_ATTRIBUTE+"=\""+toDir+"\" "+CWConstants.TO_FILE_ATTRIBUTE+"=\""+toFile+" "+CWConstants.KEY_SIZE_ATTRIBUTE+"=\"+crypto.keySize+\" "+CWConstants.PASS_PHRASE_ATTRIBUTE+"=\""+crypto.passPhrase+"\">"+LINE_SEPARATOR);
      
      al.add("  "+file+LINE_SEPARATOR);
      al.add("</"+CWConstants.ENCRYPT_FILE_ELEMENT+">"+LINE_SEPARATOR);
      
      return al;
   }
   
   /**
    * Retrieve the values in this action for the status update.
    * @return s
    */
   public String toStringForStatusUpdate(String event)
   {
      String s;

      if (toDir == null)
         s = CWConstants.ENCRYPT_FILE_ELEMENT+", "+labelName+", Encryption, "+CWConstants.CREATE_DIR_ATTRIBUTE+"="+createDir+", srcFilename="+srcFilename+", "+CWConstants.TO_FILE_ATTRIBUTE+"="+toFile+LINE_SEPARATOR;
      else
         s = CWConstants.ENCRYPT_FILE_ELEMENT+", "+labelName+", Encryption, "+CWConstants.CREATE_DIR_ATTRIBUTE+"="+createDir+", srcFilename="+srcFilename+", "+CWConstants.TO_DIR_ATTRIBUTE+"="+toDir+", "+CWConstants.TO_FILE_ATTRIBUTE+"="+toFile+LINE_SEPARATOR;

      return s;
   }
}
