package com.capitalware.ufm.actions;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.security.InvalidAlgorithmParameterException;
import java.security.InvalidKeyException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.NoSuchPaddingException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.DecryptFileObject;
import com.capitalware.ufm.handlers.AES;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.Tokens;

/**
 * This class will handle the DecryptFile action.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class DecryptFileAction extends UFMAction
{
   private DecryptFileObject   dfo;

   /**
    * The constructor
    * @param dfo
    */
   public DecryptFileAction(DecryptFileObject dfo)
   {
      super(dfo.actionName);
      this.dfo = dfo;
   }

   /**
    * Perform the sort work.
    * @return exitCode
    */
   public int doWork()
   {
      int exitCode = 0;
      String toPathAndFilename;
      dfo.printValues();

      toPathAndFilename = IO.apendFileToPath(dfo.toDir, dfo.toFile);
      toPathAndFilename = Tokens.replace(dfo.actionName, toPathAndFilename, IO.getFileName(dfo.srcFilename), 1, dfo.paClass.actionProperties);

      FileInputStream fis = null;
      FileOutputStream fos = null;
      
      try
      {
         fis = new FileInputStream(dfo.srcFilename);
         fos = new FileOutputStream(toPathAndFilename);
         AES aes = new AES(dfo.crypto.passPhrase, dfo.crypto.keySize);
         aes.decrypt(fis, fos);
         dfo.fileCount++;

         UFM.logger.info("Successfully decrypted '" + dfo.srcFilename + "' to file '" + toPathAndFilename + "'");
      }
      catch (InvalidKeyException e)
      {
         UFM.logger.fatal("InvalidKeyException: "+e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());
         exitCode = 8;
      }
      catch (NoSuchAlgorithmException e)
      {
         UFM.logger.fatal("NoSuchAlgorithmException: "+e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         exitCode = 8;
      }
      catch (NoSuchPaddingException e)
      {
         UFM.logger.fatal("NoSuchPaddingException: "+e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         exitCode = 8;
      }
      catch (InvalidAlgorithmParameterException e)
      {
         UFM.logger.fatal("InvalidAlgorithmParameterException: "+e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         exitCode = 8;
      }
      catch (UnsupportedEncodingException e)
      {
         UFM.logger.fatal("UnsupportedEncodingException: "+e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         exitCode = 8;
      }
      catch (FileNotFoundException e)
      {
         UFM.logger.fatal("FileNotFoundException: "+e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         exitCode = 8;
      }
      catch (IOException e)
      {
         UFM.logger.fatal("IOException: "+e.getLocalizedMessage());
         if (e.getCause() != null)
            UFM.logger.fatal(e.getCause().toString());;
         exitCode = 8;
      }
      finally
      {
         try
         {
            if (fis != null)
               fis.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: "+e.getLocalizedMessage());
            if (e.getCause() != null)
               UFM.logger.fatal(e.getCause().toString());;
            exitCode = 8;
         }

         try
         {
            if (fos != null)
               fos.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: "+e.getLocalizedMessage());
            if (e.getCause() != null)
               UFM.logger.fatal(e.getCause().toString());;
            exitCode = 8;
         }

         if (!dfo.onErrorFail)
            exitCode = 0; /* reset back to zero */
      }

      return exitCode;
   }
}
