package com.capitalware.ufm.handlers;

import java.io.File;
import java.io.IOException;
import java.text.DecimalFormat;
import java.util.Hashtable;

import org.jdom.JDOMException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.MQReceiveObject;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.xml.EmbeddedMsgXml;

import com.ibm.mq.MQException;
import com.ibm.mq.MQMessage;
import com.ibm.mq.constants.CMQC;

/**
 * A class to handle remote values in the MQMD or as an XML data in the message payload.
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class RemoteValuesInMsg
{
   private int          xmlDataLength = 0;
   private String       jobFileInMsg = null;
   private String       directoryInMsg = null;
   private String       filenameInMsg = null;
   
   private boolean      _runJob = false;
   private String       _jobXmlFile = null;
   private String       _toPathAndFilename = null;
   private String       _directory = null;

   /**
    * The constructor
    * @param logger
    */
   public RemoteValuesInMsg()
   {
      super();
   }

   /**
    * Get the remote values from the message.
    * @param ro
    * @param receiveMsg
    * @param msgCounter
    * @throws IOException 
    * @throws JDOMException 
    */
   public int processValues(MQReceiveObject ro, MQMessage receiveMsg, int msgCounter, Hashtable<String,String> properties) throws JDOMException, IOException
   {
      DecimalFormat df0 = new java.text.DecimalFormat("0000");
      boolean ojn = ro.overrideJobName;
      boolean odn = ro.overrideDirectoryName;
      boolean ofn = ro.overrideFileName;
      _directory = ro.defaultDirectory;
      /*
       * Set default values.
       */
      _runJob = false;
      _jobXmlFile = ro.defaultJobXmlFile;
      _toPathAndFilename = null;
      xmlDataLength = 0;
      jobFileInMsg = null;
      directoryInMsg = null;
      filenameInMsg = null;

      /* Is the message one of UFM's? */
      if ((receiveMsg.feedback & CWConstants.UFM_MSG) == CWConstants.UFM_MSG)
         extractRemoteValues(receiveMsg);
      else
      {
         /* Cannot use values from MQMD */
         ojn = true;
         odn = true;
         ofn = true;
      }

      if (ojn)
      {
         if (ro.defaultJobXmlFile != null)
            _runJob = true;
      }
      else
      {
         _jobXmlFile = jobFileInMsg;
         if ( !("".equals(_jobXmlFile)) && (_jobXmlFile != null) )
            _runJob = true;
      }

      if (!odn)
         _directory = directoryInMsg;

      /**
       * Tokenize just the directory so that the directory value will be consistent
       * for both _directory &  _toPathAndFilename just in case the user uses the 
       * date token.
       */
      _directory = Tokens.replace(_directory, properties);
      
      if (ofn)
      {
         if ( (ro.defaultFilename != null) && (!("".equals(ro.defaultFilename))) )
            _toPathAndFilename = IO.apendFileToPath(_directory, ro.defaultFilename);
         else
            _toPathAndFilename = IO.apendFileToPath(_directory, UFM.WHO_AM_I + "_"+df0.format(msgCounter+1) + ".txt");
      }
      else
         _toPathAndFilename= IO.apendFileToPath(_directory, filenameInMsg);

      _toPathAndFilename = Tokens.replace(ro.actionName, _toPathAndFilename, null, msgCounter+1, properties);

      if (_runJob)
      {
         _jobXmlFile = Tokens.replace(ro.actionName, _jobXmlFile, properties);
      }

      return 0;
   }

   /**
    * Extract remote values set by the send action.
    * @param receiveMsg
    * @throws IOException
    * @throws JDOMException
    */
   private void extractRemoteValues(MQMessage receiveMsg) throws IOException, JDOMException
   {
      if ((receiveMsg.feedback & CWConstants.UFM_V7_PROPERTIES_MSG) == CWConstants.UFM_V7_PROPERTIES_MSG)
      {
         UFM.logger.debug("RemoteValuesInMsg: getting values from properties");
         
         String tempDirectory = null;
         String tempFilename = null;
         String tempJobfile = null;
         
         try
         {
            tempDirectory = receiveMsg.getStringProperty(CWConstants.UFM_PROP_DIRECTORY);
         }
         catch (MQException e)
         {
            if ((e.completionCode == CMQC.MQCC_FAILED) && (e.reasonCode == CMQC.MQRC_PROPERTY_NOT_AVAILABLE))
               UFM.logger.debug("RemoteValuesInMsg: extractRemoteValues: '" + CWConstants.UFM_PROP_DIRECTORY + "' not found.");
            else
               UFM.logger.warn("RemoteValuesInMsg: extractRemoteValues: warning [D] MQCC=" + e.completionCode + " MQRC=" + e.reasonCode + " : "  + e.getLocalizedMessage());
         }
         
         try
         {
            tempFilename = receiveMsg.getStringProperty(CWConstants.UFM_PROP_FILENAME);
         }
         catch (MQException e)
         {
            if ((e.completionCode == CMQC.MQCC_FAILED) && (e.reasonCode == CMQC.MQRC_PROPERTY_NOT_AVAILABLE))
               UFM.logger.debug("RemoteValuesInMsg: extractRemoteValues: '" + CWConstants.UFM_PROP_FILENAME + "' not found.");
            else
               UFM.logger.warn("RemoteValuesInMsg: extractRemoteValues: warning [F] MQCC=" + e.completionCode + " MQRC=" + e.reasonCode + " : "  + e.getLocalizedMessage());
         }
         
         try
         {
            tempJobfile = receiveMsg.getStringProperty(CWConstants.UFM_PROP_EXECUTE);
         }
         catch (MQException e)
         {
            if ((e.completionCode == CMQC.MQCC_FAILED) && (e.reasonCode == CMQC.MQRC_PROPERTY_NOT_AVAILABLE))
               UFM.logger.debug("RemoteValuesInMsg: extractRemoteValues: '" + CWConstants.UFM_PROP_EXECUTE + "' not found.");
            else
               UFM.logger.warn("RemoteValuesInMsg: extractRemoteValues: warning [J] MQCC=" + e.completionCode + " MQRC=" + e.reasonCode + " : "  + e.getLocalizedMessage());
         }

         if (tempJobfile != null)
            jobFileInMsg = tempJobfile;

         directoryInMsg = tempDirectory;
         filenameInMsg = tempFilename;
      }
      else if ((receiveMsg.feedback & CWConstants.UFM_XML_EMBEDDED_MSG) == CWConstants.UFM_XML_EMBEDDED_MSG)
      {
         UFM.logger.debug("RemoteValuesInMsg: getting values from XML Msg");
         
         xmlDataLength = (int)(((receiveMsg.correlationId[0] & 0xffL) << 24) |
                               ((receiveMsg.correlationId[1] & 0xffL) << 16) |
                               ((receiveMsg.correlationId[2] & 0xffL) << 8) |
                                (receiveMsg.correlationId[3] & 0xffL));
         byte[] bx = new byte[xmlDataLength];
         
         UFM.logger.debug("bx.length="+bx.length + " : xmlDataLength=" + xmlDataLength + " : CorrelID_HEX="+StringUtilities.byteArrayToHexString(receiveMsg.correlationId));

         receiveMsg.readFully(bx);
         
         UFM.logger.debug("CorrelID_HEX="+StringUtilities.byteArrayToHexString(receiveMsg.correlationId) + "new String(bx)='" + new String(bx)+"'");

         EmbeddedMsgXml emx = new EmbeddedMsgXml();
         emx.readXMLFile(bx);
         if ((receiveMsg.feedback & CWConstants.UFM_START_JOB) == CWConstants.UFM_START_JOB)
            jobFileInMsg = emx.getJobFile();

         directoryInMsg = emx.getDirectory();
         filenameInMsg = emx.getFilename();
      }
      else
      {
         UFM.logger.debug("RemoteValuesInMsg: getting values from MQMD");
         
         if ((receiveMsg.feedback & CWConstants.UFM_START_JOB) == CWConstants.UFM_START_JOB)
            jobFileInMsg = (new String(receiveMsg.correlationId)).trim() + CWConstants.XML_SUFFIX;

         directoryInMsg = (new String(receiveMsg.accountingToken)).trim();
         filenameInMsg = (receiveMsg.applicationIdData).trim();
      }

      if ( (directoryInMsg != null) && !(directoryInMsg.endsWith(File.separator)) )
         directoryInMsg += File.separator;

      UFM.logger.info("Remote values:");
      UFM.logger.info("    JobFile             = " + (jobFileInMsg == null?"":jobFileInMsg));
      UFM.logger.info("    Directory           = " + directoryInMsg);
      UFM.logger.info("    File Name           = " + filenameInMsg);
   }

   /**
    * Run the job or not.
    * @return runJob
    */
   public boolean isRunJob()
   {
      return _runJob;
   }

   /**
    * Retrieve the Job XML File.
    * @return jobXmlFile
    */
   public String getJobXmlFile()
   {
      return _jobXmlFile;
   }

   /**
    * Retrieve the complete path and filename.
    * @return toPathAndFilename
    */
   public String getToPathAndFilename()
   {
      return _toPathAndFilename;
   }

   /**
    * Retrieve the directory
    * @return directory
    */
   public String getDirectory()
   {
      return _directory;
   }

   /**
    * Retrieve the XML data data length
    * @return xmlDataLength
    */
   public int getXmlDataLength()
   {
      return xmlDataLength;
   }
}
