package com.capitalware.ufm.handlers;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.Hashtable;
import java.util.List;

import org.jdom.JDOMException;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.internal.messages.UpdateStatus;
import com.capitalware.ufm.objects.JobObject;
import com.capitalware.ufm.xml.JobXml;

/**
 * This class will handle running an external program.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class RunJob
{
   private JobObject       job;
   private String          jobName;
   private StringBuffer    errorSB = null;
   private InternalMessageQueue    iq = null;

   /**
    * The constructor
    * @param logger
    */
   public RunJob()
   {
      super();
   }

   /**
    * The constructor
    * @param logger
    */
   public RunJob(InternalMessageQueue iq)
   {
      super();
      this.iq = iq;
   }

   /**
    * Setup and launch an external program
    * @param jobXmlFile
    * @throws IOException
    * @throws JDOMException
    */
   public void launch(String jobXmlFile, Hashtable<String,String>  properties, String actionName, String actionLabel) throws JDOMException, IOException
   {
      this.jobName = jobXmlFile;
      JobXml jxh = new JobXml();
      jxh.setParentage(actionName, actionLabel);
      jxh.setProperties(properties);
      jxh.readXMLFile(jobXmlFile);
      job = jxh.getJob();

      if (job.isWait())
      {
         executeCmd();
      }
      else
      {
         Thread t = new Thread(new Runner());
         t.start();
      }
   }

   /**
    * Setup and launch an external program
    * @param jobXmlFile
    * @param errorText
    * @throws IOException
    * @throws JDOMException
    */
   public void launch(String jobXmlFile, String errorText, Hashtable<String,String>  properties, String actionName, String actionLabel) throws JDOMException, IOException
   {
      this.jobName = jobXmlFile;
      JobXml jxh = new JobXml();
      jxh.setParentage(actionName, actionLabel);
      jxh.setProperties(properties);
      jxh.readXMLFile(jobXmlFile);
      job = jxh.getJob();
      if (errorText != null)
         errorSB = new StringBuffer(errorText);

      if (job.isWait())
      {
         executeCmd();
      }
      else
      {
         Thread t = new Thread(new Runner());
         t.start();
      }
   }

   /**
    * Common code to launch a separate external process.
    * @throws IOException
    */
   private void executeCmd() throws IOException
   {
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy/MM/dd HH:mm:ss");
      List<String> command = new ArrayList<String>();
      command.add(job.getCmd());
      String[] parms = job.getParms();
      for (int i = 0; i < parms.length; i++)
         command.add(Tokens.replace(parms[i], errorSB));

      UFM.logger.info("CMD: " + command.toString());
      
      if (iq != null)
         iq.addMessage(new UpdateStatus("0", dateTimeFormat.format(new Date())+", 0, "+job.toStringForStatusUpdate()));

      ProcessBuilder builder = new ProcessBuilder(command);
//         Map<String, String> environ = builder.environment();

      if (job.getWorkingDirectory() != null)
         builder.directory(new File(job.getWorkingDirectory()));

      Process process = builder.start();
      
      InputStream is = null;
      InputStreamReader isr = null;
      BufferedReader br = null;
      
      try
      {
         is = process.getInputStream();
         isr = new InputStreamReader(is);
         br = new BufferedReader(isr);
         String line;
         while ((line = br.readLine()) != null)
         {
            UFM.logger.info("<Output> " + line);
         }
      }
      finally
      {
         try
         {
            if (is != null)
               is.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            if (e.getCause() != null)
               UFM.logger.fatal(e.getCause().toString());
         }
            
         try
         {
            if (isr != null)
               isr.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            if (e.getCause() != null)
               UFM.logger.fatal(e.getCause().toString());
         }
         
         try
         {
            if (br != null)
               br.close();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            if (e.getCause() != null)
               UFM.logger.fatal(e.getCause().toString());
         }
      }

      UFM.logger.info("External program for '" +jobName + "' terminated.");
   }

   /**
    * This class will build and run the external in a separate thread.
    */
   class Runner implements Runnable
   {
      public void run()
      {
         UFM.logger.info("Starting a new thread to run the external program.");
         try
         {
            executeCmd();
         }
         catch (IOException e)
         {
            UFM.logger.fatal("IOException: " + e.getLocalizedMessage());
            if (e.getCause() != null)
               UFM.logger.fatal(e.getCause().toString());
         }
      }
   }
}
