package com.capitalware.ufm.handlers;

import java.text.DecimalFormat;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Date;
import java.util.Enumeration;
import java.util.Hashtable;

import com.capitalware.ufm.objects.CWConstants;

/**
 * This class will handle token replacement in a text string.
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class Tokens
{
   private static final String    DATE_TOKEN             = "${DATE}";
   private static final String    DATE_TIME_TOKEN        = "${DATETIME}";

   private static final String    DATE_TOKEN_M           = "${DATEM}";
   private static final String    USER_DATE_TOKEN_MINUS  = "${DATEMINUS}";
   private static final String[]  DATE_TOKEN_MINUS       = {"${DATE-1}", "${DATE-2}", "${DATE-3}", "${DATE-4}",  "${DATE-5}",  "${DATE-6}", "${DATE-7}" };

   private static final String    DATE_TOKEN_P           = "${DATEP}";
   private static final String    USER_DATE_TOKEN_PLUS   = "${DATEPLUS}";
   private static final String[]  DATE_TOKEN_PLUS        = {"${DATE+1}", "${DATE+2}", "${DATE+3}", "${DATE+4}",  "${DATE+5}",  "${DATE+6}", "${DATE+7}" };

   private static final String    TIME_TOKEN             = "${TIME}";
   private static final String    YEAR_TOKEN             = "${YYYY}";
   private static final String    MONTH_TOKEN            = "${MM}";
   private static final String    DAY_TOKEN              = "${DD}";
   private static final String    HOUR_TOKEN             = "${HH}";
   private static final String    HALF_HOUR_TOKEN        = "${hh}";
   private static final String    MINUTE_TOKEN           = "${MIN}";
   private static final String    SECOND_TOKEN           = "${SS}";
   //
   private static final String    FILE_TOKEN             = "${FILE}";
   private static final String    FILE_PREFIX_TOKEN      = "${FILE_PREFIX}";
   private static final String    FILE_EXTENSION_TOKEN   = "${FILE_EXT}";
   //
   private static final String    INDEX_TOKEN            = "${INDEX}";
   private static final String    ACTION_TOKEN           = "${ACTION}";
   private static final String    UFM_ERROR_TEXT_TOKEN  = "${UFM_ERROR_TEXT}";

   /**
    * The constructor - make sure nothing instantiates this class.
    * It must be accessed in a static way.
    */
   private Tokens()
   {}

   /**
    * Replace the tokens with the appropriate values.
    * @param inputText
    * @param errorSB
    * @return updatedText
    */
   public static String replace(String inputText, StringBuffer errorSB)
   {
      String target = inputText;

      if (target.indexOf('$') != -1)
      {
         if ( (errorSB != null) && (target.indexOf(UFM_ERROR_TEXT_TOKEN) != -1) )
            target = StringUtilities.replaceAll(target, UFM_ERROR_TEXT_TOKEN, errorSB.toString());
      }

      return target;
   }

   /**
    * Replace the tokens with the appropriate values.
    * @param actionName
    * @param inputText
    * @param fromFileName
    * @return updatedText
    */
   public static String replace(String actionName, String inputText, String fromFileName)
   {
      String target = inputText;

      if (target.indexOf('$') != -1)
      {
         target = replaceFile(target, fromFileName);

         target = replaceUFM(actionName, target);
      }

      return target;
   }

   /**
    * Replace the tokens with the appropriate values.
    * @param actionName
    * @param inputText
    * @param fromFileName
    * @param index
    * @param userProperties
    * @return updatedText
    */
   public static String replace(String actionName, String inputText, String fromFileName, int index, Hashtable<String,String> userProperties)
   {
      DecimalFormat df0 = new java.text.DecimalFormat("00000000");
      String target = inputText;

      if (target.indexOf('$') != -1)
      {
         if (fromFileName != null)
            target = replaceFile(target, fromFileName);

         target = replace(target, userProperties);  /* call replace(t,p) */

         target = replaceUFM(actionName, target);

         if (target.indexOf(INDEX_TOKEN) != -1)
            target = StringUtilities.replaceAll(target, INDEX_TOKEN, df0.format(index));
      }

      return target;
   }

   /**
    * Replace the tokens with the appropriate values.
    * @param actionName
    * @param inputText
    * @param userProperties
    * @return updatedText
    */
   public static String replace(String actionName, String inputText, Hashtable<String,String> userProperties)
   {
      String target = inputText;

      if (target.indexOf('$') != -1)
      {
         target = replace(target, userProperties);  /* call replace(t,p) */

         target = replaceUFM(actionName, target);
      }

      return target;
   }

   /**
    * Replace the tokens with the appropriate values.
    * @param inputText
    * @param userProperties
    * @return updatedText
    */
   public static String replace(String inputText, Hashtable<String,String> userProperties)
   {
      String target = inputText;

      if (target.indexOf('$') != -1)
      {
         if (userProperties != null)
         {
            if (!(userProperties.isEmpty()))
            {
               /* Special case - always do Loop Counter first */
               if (userProperties.containsKey(CWConstants.LOOP_COUNTER))
               {
                  if (target.indexOf(CWConstants.LOOP_COUNTER) != -1)
                     target = StringUtilities.replaceAll(target, CWConstants.LOOP_COUNTER, userProperties.get(CWConstants.LOOP_COUNTER));
               }
               
               Enumeration<String> names = userProperties.keys();

               while (names.hasMoreElements())
               {
                  String name = names.nextElement();
                  if (target.indexOf(name) != -1)
                     target = StringUtilities.replaceAll(target, name, userProperties.get(name));
               }
            }
         }

         target = replaceDateAndTime(target, userProperties);
      }

      return target;
   }

   /**
    * Replace the tokens with the appropriate values.
    * @param inputText
    * @return updatedText
    */
   private static String replaceDateAndTime(String inputText, Hashtable<String,String> userProperties)
   {
      SimpleDateFormat dateFormat     = new SimpleDateFormat("yyyy_MM_dd");
      SimpleDateFormat dateTimeFormat = new SimpleDateFormat("yyyy_MM_dd_HH_mm_ss.SSS");
      SimpleDateFormat timeFormat     = new SimpleDateFormat("HH_mm_ss.SSS");
      SimpleDateFormat yearFormat     = new SimpleDateFormat("yyyy");
      SimpleDateFormat monthFormat    = new SimpleDateFormat("MM");
      SimpleDateFormat dayFormat      = new SimpleDateFormat("dd");
      SimpleDateFormat hourFormat     = new SimpleDateFormat("HH");
      SimpleDateFormat minuteFormat   = new SimpleDateFormat("mm");
      SimpleDateFormat secondFormat   = new SimpleDateFormat("ss");
      Date now = new Date();
      Calendar cal;
      String target = inputText;

      if (target.indexOf(DATE_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, DATE_TOKEN, dateFormat.format(now));

      if (target.indexOf(DATE_TIME_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, DATE_TIME_TOKEN, dateTimeFormat.format(now));
      
      if (target.indexOf(DATE_TOKEN_M) != -1)
      {
         cal = Calendar.getInstance();
         if(userProperties.containsKey(USER_DATE_TOKEN_MINUS))
            cal.add(Calendar.DATE, -1 * StringUtilities.stringToInt(userProperties.get(USER_DATE_TOKEN_MINUS), 0));
         
         target = StringUtilities.replaceAll(target, DATE_TOKEN_M, dateFormat.format(cal.getTime()));
      }
      
      for (int i=0; i < DATE_TOKEN_MINUS.length; i++)
      {
         if (target.indexOf(DATE_TOKEN_MINUS[i]) != -1)
         {
            cal = Calendar.getInstance();
            cal.add(Calendar.DATE, -(i+1));
            target = StringUtilities.replaceAll(target, DATE_TOKEN_MINUS[i], dateFormat.format(cal.getTime()));
         }
      }
      
      if (target.indexOf(DATE_TOKEN_P) != -1)
      {
         cal = Calendar.getInstance();
         if(userProperties.containsKey(USER_DATE_TOKEN_PLUS))
            cal.add(Calendar.DATE, StringUtilities.stringToInt(userProperties.get(USER_DATE_TOKEN_PLUS), 0));
         
         target = StringUtilities.replaceAll(target, DATE_TOKEN_P, dateFormat.format(cal.getTime()));
      }

      for (int i=0; i < DATE_TOKEN_PLUS.length; i++)
      {
         if (target.indexOf(DATE_TOKEN_PLUS[i]) != -1)
         {
            cal = Calendar.getInstance();
            cal.add(Calendar.DATE, +(i+1));
            target = StringUtilities.replaceAll(target, DATE_TOKEN_PLUS[i], dateFormat.format(cal.getTime()));
         }
      }

      if (target.indexOf(TIME_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, TIME_TOKEN, timeFormat.format(now));

      if (target.indexOf(YEAR_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, YEAR_TOKEN, yearFormat.format(now));

      if (target.indexOf(MONTH_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, MONTH_TOKEN, monthFormat.format(now));

      if (target.indexOf(DAY_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, DAY_TOKEN, dayFormat.format(now));

      if (target.indexOf(HOUR_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, HOUR_TOKEN, hourFormat.format(now));

      if (target.indexOf(HALF_HOUR_TOKEN) != -1)
      {
         cal = Calendar.getInstance();
         if(cal.get(Calendar.MINUTE) >= 30)
            target = StringUtilities.replaceAll(target, HALF_HOUR_TOKEN, "30");
         else
            target = StringUtilities.replaceAll(target, HALF_HOUR_TOKEN, "00");
      }
      
      if (target.indexOf(MINUTE_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, MINUTE_TOKEN, minuteFormat.format(now));

      if (target.indexOf(SECOND_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, SECOND_TOKEN, secondFormat.format(now));

      return target;
   }

   /**
    * Replace the tokens with the appropriate values.
    * @param actionName
    * @param inputText
    * @return updatedText
    */
   private static String replaceUFM(String actionName, String inputText)
   {
      String target = inputText;

      if (target.indexOf(ACTION_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, ACTION_TOKEN, actionName);

      return target;
   }

   /**
    * Replace the tokens with the appropriate values.
    * @param inputText
    * @param fromFileName
    * @return updatedText
    */
   private static String replaceFile(String inputText, String fromFileName)
   {
      String target = inputText;
      String filePrefix = "";
      String fileExtension = "";

      if ( fromFileName != null)
      {
         int x = fromFileName.lastIndexOf('.');
         if (x != -1)
         {
            filePrefix   = fromFileName.substring(0,x);
            fileExtension = fromFileName.substring(x + 1);
         }
         else
            filePrefix   = fromFileName;
      }

      if (target.indexOf(FILE_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, FILE_TOKEN, fromFileName);

      if (target.indexOf(FILE_PREFIX_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, FILE_PREFIX_TOKEN, filePrefix);

      if (target.indexOf(FILE_EXTENSION_TOKEN) != -1)
         target = StringUtilities.replaceAll(target, FILE_EXTENSION_TOKEN, fileExtension);

      return target;
   }
}
