package com.capitalware.ufm.schedule;

import java.util.Collections;
import java.util.Date;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;
import java.util.TreeSet;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.objects.PastDateException;

/**
 * This class implements an schedule manager similar to Unix <code>cron</code> and
 * <code>at</code> daemons. It is intended to fire events when events' date and
 * time match the current ones. Schedules are added dynamically and can be one-shot
 * or repetitive (i.e. rescheduled when matched). Time unit is seconds. Events
 * scheduled less than one second to the current time are rejected (a
 * <code>PastDateException</code> is thrown).
 * <p>
 * 
 * The schedule scheduler has been designed to manage a large quantity of events
 * (it uses a priority queue to optimize schedule dates selection) and to reduce
 * the use of the CPU time (the ScheduleManager's thread is started only when there
 * are events to be managed and it sleeps until the next schedule date).
 * <p>
 * 
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class ScheduleManager
{
   protected ScheduleWaiter   waiter;
   protected SortedSet        queue; /* of EventEntry */

   /**
    * Creates a new ScheduleManager. The waiter thread will be started only when
    * the first schedule listener will be added.
    * 
    * @param isDaemon
    *           true if the waiter thread should run as a daemon.
    * @param threadName
    *           the name of the waiter thread
    */
   public ScheduleManager(boolean isDaemon, String threadName)
   {
      queue = Collections.synchronizedSortedSet(new TreeSet());

      waiter = new ScheduleWaiter(this, isDaemon, threadName);
   }

   /**
    * Creates a new ScheduleManager. The waiter thread will be started only when
    * the first schedule listener will be added. The waiter thread will <i>not</i>
    * run as a daemon.
    */
   public ScheduleManager()
   {
      this(false, "ScheduleManager");
   }

   /**
    * Adds an schedule for a specified date.
    * 
    * @param date
    *           the schedule date to be added.
    * @param listener
    *           the schedule listener.
    * @param file
    *           UFM Workflow XML file associated with this entry           
    * @return the EventEntry.
    * @exception PastDateException
    *               if the schedule date is in the past or less than 1 second
    *               closed to the current date).
    */
   public synchronized ScheduleEntry addSchedule(String _name, 
                                                 Date _date,
                                                 String _file,
                                                 ScheduleListener _listener) 
      throws PastDateException
   {
      ScheduleEntry entry = new ScheduleEntry(_name, _date, _file, _listener);
      addSchedule(entry);
      return entry;
   }

   /**
    * Adds an schedule for a specified delay.
    * 
    * @param delay
    *           the schedule delay in minute (relative to now).
    * @param isRepeating
    *           <code>true</code> if the schedule must be reactivated,
    *           <code>false</code> otherwise.
    * @param file
    *           UFM Workflow XML file associated with this entry           
    * @param listener
    *           the schedule listener.
    * @return the EventEntry.
    * @exception PastDateException
    *               if the schedule date is in the past (or less than 1 second
    *               closed to the current date).
    */
   public synchronized ScheduleEntry addSchedule(String _name, 
                                                 int _delay, 
                                                 boolean _isRepeating,
                                                 String _file,
                                                 ScheduleListener _listener)
      throws PastDateException
   {
      ScheduleEntry entry = new ScheduleEntry(_name, _delay, _isRepeating, _file, _listener);
      addSchedule(entry);
      return entry;
   }

   /**
    * Adds an schedule for a specified date.
    * 
    * @param minute
    *           minute of the event. Allowed values 0-59, or -1 for all.
    * @param hour
    *           hour of the event. Allowed values 0-23, or -1 for all.
    * @param dayOfMonth
    *           day of week of the event. Allowed values 1-31, or -1 for all.
    * @param month
    *           month of the event. Allowed values 0-11 (0 = January, 1 =
    *           February, ...), or -1 for all. <code>java.util.Calendar</code>
    *           constants can be used.
    * @param dayOfWeek
    *           day of month of the event. Allowed values 1-7 (1 = Sunday, 2 =
    *           Monday, ...), or -1 for all. <code>java.util.Calendar</code>
    *           constants can be used.
    * @param year
    *           year of the event. When this field is not set (i.e. -1) the
    *           schedule is repetitive (i.e. it is rescheduled when reached).
    * @param file
    *           UFM Workflow XML file associated with this entry           
    * @param listener
    *           the schedule listener.
    * @return the EventEntry.
    * @exception PastDateException
    *               if the schedule date is in the past (or less than 1 second away
    *               from the current date).
    */
   public synchronized ScheduleEntry addSchedule(String _name, 
                                                 int _minute,
                                                 int _hour, 
                                                 int _dayOfMonth, 
                                                 int _month, 
                                                 int _dayOfWeek, 
                                                 int _year,
                                                 String _file,
                                                 ScheduleListener _listener) 
      throws PastDateException
   {

      ScheduleEntry entry = new ScheduleEntry(_name, 
                                              _minute, 
                                              _hour, 
                                              _dayOfMonth,
                                              _month, 
                                              _dayOfWeek, 
                                              _year, 
                                              _file,
                                              _listener);
      addSchedule(entry);
      return entry;
   }

   /**
    * Adds an schedule for a specified date or matching dates (for unrestricted
    * fields).
    * 
    * @param minutes
    *           minutes of the event. Allowed values 0-59, or -1 for all.
    * @param hours
    *           hours of the event. Allowed values 0-23, or -1 for all.
    * @param daysOfMonth
    *           days of week of the event. Allowed values 1-31, or -1 for all.
    * @param months
    *           months of the event. Allowed values 0-11 (0 = January, 1 =
    *           February, ...), or -1 for all. <code>java.util.Calendar</code>
    *           constants can be used.
    * @param daysOfWeek
    *           days of month of the event. Allowed values 1-7 (1 = Sunday, 2 =
    *           Monday, ...), or -1 for all. <code>java.util.Calendar</code>
    *           constants can be used.
    * @param year
    *           year of the event. When this field is not set (i.e. -1) the
    *           schedule is repetitive (i.e. it is rescheduled when reached).
    * @param file
    *           UFM Workflow XML file associated with this entry           
    * @param listener
    *           the schedule listener.
    * @return the EventEntry.
    * @exception PastDateException
    *               if the schedule date is in the past (or less than 1 second away
    *               from the current date).
    */
   public synchronized ScheduleEntry addSchedule(String  _name, 
                                                 int[]   _minutes,
                                                 int[]   _hours, 
                                                 int[]   _daysOfMonth, 
                                                 int[]   _months, 
                                                 int[]   _daysOfWeek,
                                                 int     _year, 
                                                 String  _file, 
                                                 ScheduleListener _listener) 
   throws PastDateException
   {
      ScheduleEntry entry = new ScheduleEntry(_name, _minutes, _hours, _daysOfMonth, _months, _daysOfWeek, _year, _file, _listener);
      addSchedule(entry);
      return entry;
   }

   /**
    * Adds an schedule for a specified EventEntry
    * 
    * @param entry
    *           the EventEntry.
    * @exception PastDateException
    *               if the schedule date is in the past (or less than one second
    *               away from the current date).
    */
   public synchronized void addSchedule(ScheduleEntry _entry) // throws PastDateException
   {
      UFM.logger.debug("Add a new schedule entry : " + _entry);

      boolean flag = queue.add(_entry);

      if (flag == false)
      {
         System.err.println(_entry.getName() + " already existed.");
      }

      if (queue.first().equals(_entry))
      {
         UFM.logger.debug("This new schedule is the top one, update the waiter thread");
         waiter.update(_entry.eventTime);
      }
   }

   /**
    * Removes the specified EventEntry.
    * 
    * @param entry
    *           the EventEntry that needs to be removed.
    * @return <code>true</code> if there was an schedule for this date,
    *         <code>false</code> otherwise.
    */
   public synchronized boolean removeSchedule(ScheduleEntry _entry)
   {
      boolean found = false;

      if (!queue.isEmpty())
      {
         ScheduleEntry was_first = (ScheduleEntry) queue.first();
         found = queue.remove(_entry);

         // update the queue if it's not now empty, and the first schedule has
         // changed
         if (!queue.isEmpty() && _entry.equals(was_first))
         {
            waiter.update(((ScheduleEntry) queue.first()).eventTime);
         }
      }

      return found;
   } // removeEvent()

   /**
    * Removes all the events. No more events, even newly added ones, will be
    * fired.
    */
   public synchronized void removeAllSchedules()
   {
      queue.clear();
   }

   /**
    * Removes all the events. No more events, even newly added ones, will be
    * fired.
    */
   public synchronized void removeAllSchedulesAndStop()
   {
      waiter.stopIt();
      waiter = null;
      queue.clear();
   }

   public boolean isStopped()
   {
      return (waiter == null);
   }

   /**
    * Tests whether the supplied EventEntry is in the manager.
    * 
    * @param ScheduleEntry
    * @return boolean whether EventEntry is contained within the manager
    */
   public synchronized boolean containsSchedule(ScheduleEntry _eventEntry)
   {
      return queue.contains(_eventEntry);
   }

   /**
    * Returns a copy of all events in the manager.
    */
   public synchronized List getAllSchedules()
   {
      List result = new ArrayList();

      Iterator iterator = queue.iterator();
      while (iterator.hasNext())
      {
         result.add(iterator.next());
      }

      return result;
   }

   /**
    * This is method is called when an schedule date is reached. It is only be
    * called by the the EventWaiter or by itself (if the next schedule is less than
    * 1 second away).
    */
   protected synchronized void notifyNextSchedule()
   {
      UFM.logger.debug("notifying next event");

      // if the queue is empty, there's nothing to do
      if (queue.isEmpty())
      {
         UFM.logger.debug("queue is empty");
         return;
      }

      // Removes this schedule and notifies the listener
      ScheduleEntry entry = (ScheduleEntry) queue.first();
      // debug("entry : " + entry.getName());
      queue.remove(entry);

      // NOTE: if the entry is still running when its next schedule time comes up,
      // that execution of the entry will be skipped.
      if (entry.isNotificationInNewThread())
      {
         // debug("new Thread");
         new Thread(new RunnableRinger(entry)).start();
      }
      else
      {
         // ring in same thread, sequentially.. can delay other events
         try
         {
            // debug("entry.eventNotification()");
            entry.scheduleNotification();
         }
         catch (Exception e)
         {
            e.printStackTrace();
         }
      }

      // Reactivates the schedule if it is repetitive
      if (entry.isRepeating)
      {
         // debug("entry.isRepeating : " + entry.getName());
         // Iterator iterator = queue.iterator();
         // while (iterator.hasNext())
         // {
         // System.err.println("1: "+iterator.next());
         // }

         entry.updateScheduleTime();
         boolean flag = queue.add(entry);

         UFM.logger.debug("entry.isRepeating : " + entry.getName() + " : flag=" + flag);

         if (flag == false)
         {
            Iterator iterator = queue.iterator();
            while (iterator.hasNext())
            {
               System.err.println("2: " + iterator.next());
            }
         }
      }

      // debug("queue.isEmpty()" + entry.getName());

      // Notifies the EventWaiter thread for the next event
      if (queue.isEmpty())
      {
         UFM.logger.debug("no more events to handle; queue is empty");
      }
      else
      {
         long eventTime = ((ScheduleEntry) queue.first()).eventTime;
         if (eventTime - System.currentTimeMillis() < 1000)
         {
            UFM.logger.debug("next schedule is within 1 sec or already past - ring it without waiting");
            notifyNextSchedule();
         }
         else
         {
            UFM.logger.debug("updating the waiter for next event: " + queue.first());
            waiter.restart(eventTime);
         }
      }

      UFM.logger.debug("leaving : entry : " + entry.getName());
   } // notifyListeners()

   /**
    * Stops the waiter thread before ending.
    */
   public void finalize()
   {
      if (waiter != null)
         waiter.stopIt();
   }

   /**
    * Used to ring an EventEntry in a new Thread.
    * 
    * @see com.capitalware.ufm.schedule.ScheduleEntry#setRingInNewThread()
    */
   private class RunnableRinger implements Runnable
   {
      ScheduleEntry entry = null;

      RunnableRinger(ScheduleEntry _entry)
      {
         entry = _entry;
      }

      public void run()
      {
         try
         {
            entry.scheduleNotification();
         }
         catch (Exception e)
         {
            e.printStackTrace();
         }
      }
   }
}
