package com.capitalware.ufm.xml;

import java.io.*;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.*;

import org.apache.log4j.Appender;
import org.apache.log4j.FileAppender;
import org.apache.log4j.PropertyConfigurator;
import org.apache.log4j.Logger;

import org.jdom.*;
import org.jdom.input.*;

import com.ibm.mq.constants.CMQC;

import com.capitalware.ufm.UFM;
import com.capitalware.ufm.action.objects.*;
import com.capitalware.ufm.handlers.IO;
import com.capitalware.ufm.handlers.StringUtilities;
import com.capitalware.ufm.handlers.Tokens;
import com.capitalware.ufm.mq.objects.QueueObject;
import com.capitalware.ufm.mq.objects.StatusObject;
import com.capitalware.ufm.objects.CWConstants;
import com.capitalware.ufm.objects.Cryptography;
import com.capitalware.ufm.watch.DirectoryWatch;
import com.capitalware.ufm.watch.FileWatch;
import com.capitalware.ufm.watch.Watch;
import com.capitalware.ufm.xml.DTDs.ResolveLocalDTD;

/**
 * This class handles the reading and parsing of UFM XML file and parse parameters
 *
 * @author Roger Lacroix
 * @version 1.0
 * @license Apache 2 License
 */
public class UFMXml extends UFMXmlAbstract
{
   private ArrayList<UFMObject>      _actionAL;
   private Hashtable<String,String> _properties =  new Hashtable<String,String>();
   private OnErrorObject            _onError = null;
   //
   private String                   mqXMLFile = null;
   private String                   CCDTFile = null;
   private String                   qMgrName = null;
   private QueueObject              qObject = null;
   //
   private String                   defaultMQXmlFile = null;
   private String                   defaultCCDTFile = null;
   private String                   defaultQMgrName = null;
   private QueueObject              defaultQObject = null;
   private QueueObject              defaultBackOutQObject = null;
   private QueueObject              defaultReplyToQObject = null;
   
   private boolean                  defaultOnErrorFail = true;
   
   private StatusObject             statusObject = null;
   
   private Logger                   logger;

   /**
    * The constructor
    *
    * @param logger
    */
   public UFMXml()
   {
      super();
   }

   /**
    * Set the Log4J logger
    * @param logger
    */
   public void setLogger(Logger logger)
   {
      this.logger = logger;
   }
   
   /**
    * Read xml file and set parameters
    *
    * @param xmlFilename
    *           The name of xml file
    * @throws JDOMException, IOException
    */
   public void readXMLFile(String xmlFilename) throws JDOMException, IOException
   {
      SAXBuilder builder = new SAXBuilder(true);

      // handle the ability to pull DTD from Jar if needed
      builder.setEntityResolver(new ResolveLocalDTD(CWConstants.WORKFLOW_DTD));

      Document doc = builder.build(xmlFilename);

      Element rootElement = doc.getRootElement();
      
      addServerSettings();

      if ( rootElement.getChild(CWConstants.GLOBAL_ELEMENT) != null)
         retrieveGlobal(rootElement);

      Element actionsElement = rootElement.getChild(CWConstants.ACTIONS_ELEMENT);
      
      _actionAL = extractActions(actionsElement);
   }
   
   /**
    * 
    *
    * @param xmlFilename
    *           The name of xml file
    * @throws JDOMException, IOException
    */
   public ArrayList<UFMObject> extractActions(Element actionsElement) throws JDOMException, IOException
   {
      ArrayList<UFMObject> al = new ArrayList<UFMObject>();

      Iterator it = actionsElement.getChildren().iterator();

      while (it.hasNext())
      {
         Element actionElement = (Element) it.next();
         String actionTagName = actionElement.getName();

         if (CWConstants.MQ_SEND_ELEMENT.equals(actionTagName))
            al.add(retrieveMQSend(actionElement));
         else if  (CWConstants.MQ_RECEIVE_ELEMENT.equals(actionTagName))
            al.add(retrieveMQReceive(actionElement));
         else if  (CWConstants.MQ_PUTQUIT_ELEMENT.equals(actionTagName))
            al.add(retrieveMQPutQuit(actionElement));
         else if  (CWConstants.ECHO_ELEMENT.equals(actionTagName))
            al.add(retrieveEcho(actionElement));
         else if  (CWConstants.EXECUTE_ELEMENT.equals(actionTagName))
            al.add(retrieveExecute(actionElement));
         else if  (CWConstants.FTP_ELEMENT.equals(actionTagName))
            al.add(retrieveFtp(actionElement));
         else if  (CWConstants.SFTP_ELEMENT.equals(actionTagName))
            al.add(retrieveSFtp(actionElement));
         else if  (CWConstants.IF_ELEMENT.equals(actionTagName))
            al.add(retrieveIf(actionElement));
         else if  (CWConstants.HTTP_GET_FILE_ELEMENT.equals(actionTagName))
            al.add(retrieveHttpGetFile(actionElement));
         else if  (CWConstants.HTTP_PUT_FILE_ELEMENT.equals(actionTagName))
            al.add(retrieveHttpPutFile(actionElement));
         else if  (CWConstants.LOOP_ELEMENT.equals(actionTagName))
            al.add(retrieveLoop(actionElement));
         else if  (CWConstants.LAUNCH_ELEMENT.equals(actionTagName))
            al.add(retrieveLaunch(actionElement));
         else if  (CWConstants.SCHEDULE_ELEMENT.equals(actionTagName))
            al.add(retrieveSchedule(actionElement));
         else if  (CWConstants.SEND_EMAIL_ELEMENT.equals(actionTagName))
            al.add(retrieveSendEmail(actionElement));
         else if  (CWConstants.SLEEP_ELEMENT.equals(actionTagName))
            al.add(retrieveSleep(actionElement));
         else if  (CWConstants.APPEND_ELEMENT.equals(actionTagName))
            al.add(retrieveAppend(actionElement));
         else if  (CWConstants.CONVERT_ELEMENT.equals(actionTagName))
            al.add(retrieveConvert(actionElement));
         else if  (CWConstants.COPY_ELEMENT.equals(actionTagName))
            al.add(retrieveCopy(actionElement));
         else if  (CWConstants.DECRYPT_FILE_ELEMENT.equals(actionTagName))
            al.add(retrieveDecryptFile(actionElement));
         else if  (CWConstants.DELETE_ELEMENT.equals(actionTagName))
            al.add(retrieveDelete(actionElement));
         else if  (CWConstants.ENCRYPT_FILE_ELEMENT.equals(actionTagName))
            al.add(retrieveEncryptFile(actionElement));
         else if  (CWConstants.MAKEDIR_ELEMENT.equals(actionTagName))
            al.add(retrieveMakeDir(actionElement));
         else if  (CWConstants.MERGE_ELEMENT.equals(actionTagName))
            al.add(retrieveMerge(actionElement));
         else if  (CWConstants.MERGE_SORT_ELEMENT.equals(actionTagName))
            al.add(retrieveMergeSort(actionElement));
         else if  (CWConstants.MOVE_ELEMENT.equals(actionTagName))
            al.add(retrieveMove(actionElement));
         else if  (CWConstants.REENCODE_ELEMENT.equals(actionTagName))
            al.add(retrieveReEncode(actionElement));
         else if  (CWConstants.REMOVE_ITEMS_ELEMENT.equals(actionTagName))
            al.add(retrieveRemoveItems(actionElement));
         else if  (CWConstants.REMOVE_DIR_ELEMENT.equals(actionTagName))
            al.add(retrieveRemoveDir(actionElement));
         else if  (CWConstants.RENAME_ELEMENT.equals(actionTagName))
            al.add(retrieveRename(actionElement));
         else if  (CWConstants.REPLACE_TEXT_ELEMENT.equals(actionTagName))
            al.add(retrieveReplaceText(actionElement));
         else if  (CWConstants.SCP_ELEMENT.equals(actionTagName))
            al.add(retrieveScp(actionElement));
         else if  (CWConstants.SORT_ELEMENT.equals(actionTagName))
            al.add(retrieveSort(actionElement));
         else if  (CWConstants.SORT_UNIQUE_ELEMENT.equals(actionTagName))
            al.add(retrieveSortUnique(actionElement));
         else if  (CWConstants.SPLIT_ELEMENT.equals(actionTagName))
            al.add(retrieveSplit(actionElement));
         else if  (CWConstants.TAR_ELEMENT.equals(actionTagName))
            al.add(retrieveTar(actionElement));
         else if  (CWConstants.TOUCH_ELEMENT.equals(actionTagName))
            al.add(retrieveTouch(actionElement));
         else if  (CWConstants.UNIQUE_ELEMENT.equals(actionTagName))
            al.add(retrieveUnique(actionElement));
         else if  (CWConstants.UNTAR_ELEMENT.equals(actionTagName))
            al.add(retrieveUnTar(actionElement));
         else if  (CWConstants.UNZIP_ELEMENT.equals(actionTagName))
            al.add(retrieveUnZip(actionElement));
         else if  (CWConstants.WATCH_ELEMENT.equals(actionTagName))
            al.add(retrieveWatch( actionElement));
         else if  (CWConstants.WRITE_TEXT_ELEMENT.equals(actionTagName))
            al.add(retrieveWriteText(actionElement));
         else if  (CWConstants.ZIP_ELEMENT.equals(actionTagName))
            al.add(retrieveZip(actionElement));
         else
            throw new JDOMException("Unknown Action tag '"+actionTagName+"'");
      }
      
      return al;
   }

   /**
    * Retrieve all XML values for Global tag.
    * @param rootElement
    * @throws JDOMException
    */
   private void retrieveGlobal(Element rootElement) throws JDOMException
   {
      Element gblElement = rootElement.getChild(CWConstants.GLOBAL_ELEMENT);
      Element mqElement = null;
      Element ofElement = null;
      Element execElement = null;
      Element queueElement = null;
      Element statusElement = null;
      Element statusMQElement = null;
      char o = ' ';

      /**
       * Get attributes first.
       */
      defaultOnErrorFail = getOnErrorFailAttribute(gblElement, defaultOnErrorFail);

      if (gblElement.getAttribute(CWConstants.USE_FILE_LOCKING) != null)
         o = gblElement.getAttribute(CWConstants.USE_FILE_LOCKING).getValue().toUpperCase().charAt(0);

      if (o == 'Y')
         UFM.useFileLocking = true;
      else if (o == 'N')
         UFM.useFileLocking = false;
      
      /**
       * now get elements
       */

      /* Do we have an Log4J_Properties_File element? */
      if ( gblElement.getChild(CWConstants.LOG4J_PROP_FILE_ELEMENT) != null)
      {
         String log4jPropFile = Tokens.replace(gblElement.getChild(CWConstants.LOG4J_PROP_FILE_ELEMENT).getTextTrim(), _properties);
         
         PropertyConfigurator.configure(log4jPropFile);
      }

      /* Do we have an LogFile element? */
      if ( gblElement.getChild(CWConstants.LOG_FILE_ELEMENT) != null)
      {
         String logfile = Tokens.replace(gblElement.getChild(CWConstants.LOG_FILE_ELEMENT).getTextTrim(), _properties);
         
         Appender appender = logger.getAppender(CWConstants.LOG4J_APPENDER_UFM);
         if ( logfile.indexOf(File.separator) == -1 )
            ((FileAppender) appender).setFile("log"+File.separator+logfile);
         else
            ((FileAppender) appender).setFile(logfile);

         ((FileAppender) appender).activateOptions();
      }

      /* Do we have a Property elements? */
      List propertyList = gblElement.getChildren(CWConstants.PROPERTY_ELEMENT);
      Iterator it = propertyList.iterator();
      while (it.hasNext())
      {
         String name = null;
         String value = null;
         Element propElement = (Element) it.next();

         if (propElement.getAttribute(CWConstants.NAME_ATTRIBUTE) != null)
            name = propElement.getAttribute(CWConstants.NAME_ATTRIBUTE).getValue().trim();

         if (propElement.getAttribute(CWConstants.VALUE_ATTRIBUTE) != null)
            value = propElement.getAttribute(CWConstants.VALUE_ATTRIBUTE).getValue().trim();

         if ( (name != null) && (value != null) && (!("".equals(name))) && (!("".equals(value))) )
            _properties.put(("${"+name+"}"), Tokens.replace(value, _properties));
      }

      /* Do we have an OnFailure element? */
      if ((ofElement = gblElement.getChild(CWConstants.ON_ERROR_ELEMENT)) != null)
      {
         if ((execElement = ofElement.getChild(CWConstants.EXECUTE_ELEMENT)) != null)
         {
            if (_onError == null)
               _onError = new OnErrorObject();
            
            _onError.eo = new ExecuteObject();

            if (execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
               _onError.eo.xmlFile = Tokens.replace(execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim(), _properties);
         }
         
         if ((execElement = ofElement.getChild(CWConstants.SEND_EMAIL_ELEMENT)) != null)
         {
            if (_onError == null)
               _onError = new OnErrorObject();
            
            _onError.seo = new SendEmailObject();

            if (execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
               _onError.seo.xmlFile = Tokens.replace(execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim(), _properties);
         }
      }

      /* Do we have an MQ element? */
      if ((mqElement = gblElement.getChild(CWConstants.MQ_ELEMENT)) != null)
      {
         if ( mqElement.getChild(CWConstants.MQFILE_ELEMENT) != null)
            defaultMQXmlFile = mqElement.getChild(CWConstants.MQFILE_ELEMENT).getTextTrim();

         if ( mqElement.getChild(CWConstants.CCDTFILE_ELEMENT) != null)
            defaultCCDTFile = mqElement.getChild(CWConstants.CCDTFILE_ELEMENT).getTextTrim();

         if ( (defaultMQXmlFile != null) && (defaultCCDTFile != null) )
            throw new JDOMException("Error: Do not specify both "+CWConstants.MQFILE_ELEMENT+" and "+CWConstants.CCDTFILE_ELEMENT+" elements.");

         if ( mqElement.getChild(CWConstants.QMGRNAME_ELEMENT) != null)
            defaultQMgrName = mqElement.getChild(CWConstants.QMGRNAME_ELEMENT).getTextTrim();
         
         if ( (queueElement = mqElement.getChild(CWConstants.QUEUENAME_ELEMENT)) != null)
         {
            defaultQObject = new QueueObject();
            defaultQObject.qName = queueElement.getTextTrim();

            if (queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE) != null)
               defaultQObject.qMgrName = queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE).getValue().trim();
            else
               defaultQObject.qMgrName = null;
         }
         
         if ( (queueElement = mqElement.getChild(CWConstants.BACKOUTQNAME_ELEMENT)) != null)
         {
            defaultBackOutQObject = new QueueObject();
            defaultBackOutQObject.qName = queueElement.getTextTrim();

            if (queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE) != null)
               defaultBackOutQObject.qMgrName = queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE).getValue().trim();
            else
               defaultBackOutQObject.qMgrName = null;
         }

         if ( (queueElement = mqElement.getChild(CWConstants.REPLYTOQNAME_ELEMENT)) != null)
         {
            defaultReplyToQObject = new QueueObject();
            defaultReplyToQObject.qName = queueElement.getTextTrim();

            if (queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE) != null)
               defaultReplyToQObject.qMgrName = queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE).getValue().trim();
            else
               defaultReplyToQObject.qMgrName = null;
         }
      }

      /* Do we have an Status element? */
      if ((statusElement = gblElement.getChild(CWConstants.STATUS_ELEMENT)) != null)
      {
         int timeToLive = CWConstants.TIME_TO_LIVE_DEFAULT;
         if (statusElement.getAttribute(CWConstants.TIME_TO_LIVE_ATTRIBUTE) != null)
            timeToLive = StringUtilities.stringToInt(statusElement.getAttribute(CWConstants.TIME_TO_LIVE_ATTRIBUTE).getValue().trim(), CWConstants.TIME_TO_LIVE_DEFAULT);
         
         if ((statusMQElement = statusElement.getChild(CWConstants.MQ_ELEMENT)) != null)
         {
            statusObject = new StatusObject();
            
            statusObject.timeToLive = timeToLive;
            
            /* Set default values from global values */
            statusObject.mqXMLFile= defaultMQXmlFile;
            statusObject.CCDTFile = defaultCCDTFile;
            statusObject.qMgrName = defaultQMgrName;
            statusObject.qObject  = defaultQObject;

            if ( statusMQElement.getChild(CWConstants.MQFILE_ELEMENT) != null)
               statusObject.mqXMLFile = statusMQElement.getChild(CWConstants.MQFILE_ELEMENT).getTextTrim();
            
            if ( statusMQElement.getChild(CWConstants.CCDTFILE_ELEMENT) != null)
               statusObject.CCDTFile = statusMQElement.getChild(CWConstants.CCDTFILE_ELEMENT).getTextTrim();

            if ( (statusObject.mqXMLFile != null) && (statusObject.CCDTFile != null) )
               throw new JDOMException("Error: Do not specify both "+CWConstants.MQFILE_ELEMENT+" and "+CWConstants.CCDTFILE_ELEMENT+" elements.");

            if ( statusMQElement.getChild(CWConstants.QMGRNAME_ELEMENT) != null)
               statusObject.qMgrName = statusMQElement.getChild(CWConstants.QMGRNAME_ELEMENT).getTextTrim();

            if ( (queueElement = statusMQElement.getChild(CWConstants.QUEUENAME_ELEMENT)) != null)
            {
               statusObject.qObject = new QueueObject();
               statusObject.qObject.qName = queueElement.getTextTrim();

               if (queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE) != null)
                  statusObject.qObject.qMgrName = queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE).getValue().trim();
               else
                  statusObject.qObject.qMgrName = null;
            }

            if (statusObject.mqXMLFile != null)
               statusObject.mqXMLFile  = Tokens.replace(statusObject.mqXMLFile, _properties);
            if (statusObject.CCDTFile != null)
               statusObject.CCDTFile = Tokens.replace(statusObject.CCDTFile, _properties);
            if (statusObject.qMgrName != null)
               statusObject.qMgrName = Tokens.replace(statusObject.qMgrName, _properties);
            if (statusObject.qObject != null)
            {
               if (statusObject.qObject.qName != null)
                  statusObject.qObject.qName = Tokens.replace(statusObject.qObject.qName, _properties);
               if (statusObject.qObject.qMgrName != null)
                  statusObject.qObject.qMgrName = Tokens.replace(statusObject.qObject.qMgrName, _properties);
            }
         }
      }
   }

   /**
    * Retrieve all XML values for send action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveMQSend(Element actionElement) throws JDOMException
   {
      Element mqElement = retrieveMQ(actionElement);
      
      Element fileElement = null;
      Element execElement = null;
      Element archiveElement = null;
      Element queueElement = null;
      String  flag = " ";

      MQSendObject so = new MQSendObject();
      so.mqXMLFile = mqXMLFile;
      so.CCDTFile = CCDTFile;
      so.qMgrName = qMgrName;
      so.qObject = qObject;
      so.replyToQName = defaultReplyToQObject;
      
      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.USE_V7_PROPERTIES).getValue().toUpperCase().charAt(0) == 'Y')
         so.usev7prop = true;
      else
         so.usev7prop = false;

      if (actionElement.getAttribute(CWConstants.PRIORITY_ATTRIBUTE) != null)
      {
         String priority = actionElement.getAttribute(CWConstants.PRIORITY_ATTRIBUTE).getValue().trim();
         so.priority = StringUtilities.stringToInt(priority, CMQC.MQPRI_PRIORITY_AS_Q_DEF);
      }

      so.format = actionElement.getAttribute(CWConstants.FORMAT_ATTRIBUTE).getValue().toUpperCase().charAt(0);

      if (actionElement.getAttribute(CWConstants.PERSISTENCE_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         so.persistence = true;
      else
         so.persistence = false;

      if (actionElement.getAttribute(CWConstants.CCSID_ATTRIBUTE) != null)
      {
         String ccsid = actionElement.getAttribute(CWConstants.CCSID_ATTRIBUTE).getValue().trim();
         so.ccsid = StringUtilities.stringToInt(ccsid, CMQC.MQCCSI_Q_MGR);
      }

      if (actionElement.getAttribute(CWConstants.DELETE_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         so.deleteFile = true;

      if (actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          so.crypto = new Cryptography();
          so.crypto.passPhrase = actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE).getValue().trim();

         if (actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE) != null)
         {
             String keySize = actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE).getValue().trim();
             so.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
         }
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            so.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            so.srcFilename = filename;
      }
      
      if ( (mqElement != null) && ((queueElement = mqElement.getChild(CWConstants.REPLYTOQNAME_ELEMENT)) != null) )
      {
         so.replyToQName = new QueueObject();
         so.replyToQName.qName = queueElement.getTextTrim();

         if (queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE) != null)
            so.replyToQName.qMgrName = queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE).getValue().trim();
         else
            so.replyToQName.qMgrName = null;
      }

      if ((execElement = actionElement.getChild(CWConstants.EXECUTE_ELEMENT)) != null)
      {
         if (execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
            so.jobXmlFile = execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();
      }

      if ((archiveElement = actionElement.getChild(CWConstants.ARCHIVE_ELEMENT)) != null)
      {
         if (archiveElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
            so.archiveDirectory = archiveElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

         if (archiveElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE) != null)
            so.archiveFileName = archiveElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

         if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
         {
            flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
            if (flag.charAt(0) == CWConstants.YES_FLAG)
               so.archiveCreateDir = true;
         }

         so.archiveFile = true;
      }

      if ( actionElement.getChild(CWConstants.REMOTE_ELEMENT) != null)
      {
         Element remoteElement = actionElement.getChild(CWConstants.REMOTE_ELEMENT);

         if ((execElement = remoteElement.getChild(CWConstants.EXECUTE_ELEMENT)) != null)
         {
            if (execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
               so.remoteJobXmlFile = execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();
         }

         if ( remoteElement.getChild(CWConstants.DIRECTORY_ELEMENT) != null)
            so.remoteDirectory = remoteElement.getChild(CWConstants.DIRECTORY_ELEMENT).getTextTrim();

         if ( remoteElement.getChild(CWConstants.FILENAME_ELEMENT) != null)
            so.remoteFilename = remoteElement.getChild(CWConstants.FILENAME_ELEMENT).getTextTrim();
      }

      return so;
   }

   /**
    * Retrieve all XML values for receive action.
    * @param actionElement
    * @throws JDOMException
    * @throws IOException 
    */
   private UFMObject retrieveMQReceive(Element actionElement) throws JDOMException, IOException
   {
      Element mqElement = retrieveMQ(actionElement);
      Element remoteElement = null;
      Element execElement = null;
      Element ddElement = null;
      Element dfnElement = null;
      Element queueElement = null;
      Element actionsElement;
      String  run = CWConstants.ONE_PASS + "";
      String  flag;

      MQReceiveObject ro = new MQReceiveObject();
      ro.mqXMLFile = mqXMLFile;
      ro.CCDTFile = CCDTFile;
      ro.qMgrName = qMgrName;
      ro.qObject = qObject;
      ro.backOutQName = defaultBackOutQObject;
      
      ro.labelName = getActionLabelAttribute(actionElement);

      ro.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.GETWITHCONVERT_ATTRIBUTE).getValue().toUpperCase().charAt(0) == 'Y')
         ro.getWithConvert = true;
      else
         ro.getWithConvert = false;

      if (actionElement.getAttribute(CWConstants.RUN_ATTRIBUTE) != null)
         run = actionElement.getAttribute(CWConstants.RUN_ATTRIBUTE).getValue().toUpperCase();

      if (actionElement.getAttribute(CWConstants.APPEND_ATTRIBUTE) != null)
      {
         flag = actionElement.getAttribute(CWConstants.APPEND_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            ro.append = true;
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            ro.createDir = true;
      }

      if (actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          ro.crypto = new Cryptography();
          ro.crypto.passPhrase = actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE).getValue().trim();

         if (actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE) != null)
         {
             String keySize = actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE).getValue().trim();
             ro.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
         }
      }
      
      if ( (mqElement != null) && ((queueElement = mqElement.getChild(CWConstants.BACKOUTQNAME_ELEMENT)) != null) )
      {
         ro.backOutQName = new QueueObject();         
         ro.backOutQName.qName = queueElement.getTextTrim();

         if (queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE) != null)
            ro.backOutQName.qMgrName = queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE).getValue().trim();
         else
            ro.backOutQName.qMgrName = null;
      }
      
      if ((remoteElement = actionElement.getChild(CWConstants.DEFAULT_ELEMENT)) != null)
      {
         if ((execElement = remoteElement.getChild(CWConstants.EXECUTE_ELEMENT)) != null)
         {
            if (execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
               ro.defaultJobXmlFile = execElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();
            
            if (execElement.getAttribute(CWConstants.OVERRIDE_ATTRIBUTE) != null)
            {
               String djnOverride = execElement.getAttribute(CWConstants.OVERRIDE_ATTRIBUTE).getValue().trim().toUpperCase();
               if (djnOverride.charAt(0) == CWConstants.YES_FLAG)
                  ro.overrideJobName = true;
            }
         }

         if ((ddElement = remoteElement.getChild(CWConstants.DIRECTORY_ELEMENT)) != null)
         {
            ro.defaultDirectory = ddElement.getTextTrim();
            
            if (ddElement.getAttribute(CWConstants.OVERRIDE_ATTRIBUTE) != null)
            {
               String ddOverride = ddElement.getAttribute(CWConstants.OVERRIDE_ATTRIBUTE).getValue().trim().toUpperCase();
               if (ddOverride.charAt(0) == CWConstants.YES_FLAG)
                  ro.overrideDirectoryName = true;
            }
         }

         if ((dfnElement = remoteElement.getChild(CWConstants.FILENAME_ELEMENT)) != null)
         {
            ro.defaultFilename = dfnElement.getTextTrim();
            
            if (dfnElement.getAttribute(CWConstants.OVERRIDE_ATTRIBUTE) != null)
            {
               String dfnOverride = dfnElement.getAttribute(CWConstants.OVERRIDE_ATTRIBUTE).getValue().trim().toUpperCase();
               if (dfnOverride.charAt(0) == CWConstants.YES_FLAG)
                  ro.overrideFileName = true;
            }
         }
      }

      if ( (actionsElement = actionElement.getChild(CWConstants.ACTIONS_ELEMENT)) != null)
         ro.actionAL = extractActions(actionsElement);

      if (run.charAt(0) == 'E')
         ro.processType = CWConstants.RUN_UNTIL_EMPTY;
      else if (run.charAt(0) == 'D')
         ro.processType = CWConstants.DAEMON;
      else
         ro.processType = CWConstants.ONE_PASS;

      return ro;
   }

   /**
    * Retrieve all XML values for putquit action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveMQPutQuit(Element actionElement) throws JDOMException
   {
      retrieveMQ(actionElement);

      MQPutQuitObject pqo = new MQPutQuitObject();
      pqo.mqXMLFile = mqXMLFile;
      pqo.CCDTFile = CCDTFile;
      pqo.qMgrName = qMgrName;
      pqo.qObject = qObject;
      
      pqo.labelName = getActionLabelAttribute(actionElement);

      pqo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      return pqo;
   }

   /**
    * Retrieve all XML values for Echo action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveEcho(Element actionElement)
   {
      EchoObject eo = new EchoObject();

      eo.labelName = getActionLabelAttribute(actionElement);

      eo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      eo.text = actionElement.getTextTrim();

      return eo;
   }

   /**
    * Retrieve all XML values for execute action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveExecute(Element actionElement)
   {
      ExecuteObject eo = new ExecuteObject();

      eo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
         eo.xmlFile = actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();

      return eo;
   }

   /**
    * Retrieve all XML values for Ftp action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveFtp(Element actionElement)
   {
      FtpObject fo = new FtpObject();

      fo.labelName = getActionLabelAttribute(actionElement);

      fo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
         fo.xmlFile = actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();

      return fo;
   }

   /**
    * Retrieve all XML values for Sftp action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveSFtp(Element actionElement)
   {
      SFtpObject fo = new SFtpObject();

      fo.labelName = getActionLabelAttribute(actionElement);

      fo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
         fo.xmlFile = actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();

      return fo;
   }

   /**
    * Retrieve all XML values for HttpGetFile action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveHttpGetFile(Element actionElement)
   {
      Element fileElement = null;

      HttpGetFileObject hgfo = new HttpGetFileObject();

      hgfo.labelName = getActionLabelAttribute(actionElement);

      hgfo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.URL_ATTRIBUTE) != null)
         hgfo.url = actionElement.getAttribute(CWConstants.URL_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            hgfo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         hgfo.toFile = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            hgfo.toDir = fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim();
      }

      return hgfo;
   }

   /**
    * Retrieve all XML values for HttpPutFile action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveHttpPutFile(Element actionElement)
   {
      Element fileElement = null;

      HttpPutFileObject hpfo = new HttpPutFileObject();

      hpfo.labelName = getActionLabelAttribute(actionElement);

      hpfo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      hpfo.url = actionElement.getAttribute(CWConstants.URL_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.FORM_DATA_NAME_ATTRIBUTE) != null)
         hpfo.formDataName = actionElement.getAttribute(CWConstants.FORM_DATA_NAME_ATTRIBUTE).getValue().trim();
      
      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            hpfo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            hpfo.srcFilename = filename;
      }

      return hpfo;
   }

   /**
    * Retrieve all XML values for if action.
    * @param ifElement
    * @throws JDOMException
    * @throws IOException 
    */
   private UFMObject retrieveIf(Element ifElement) throws JDOMException, IOException
   {
      Element actionsElement;
      Element elseElement;
      IfObject io = new IfObject();
      
      io.onErrorFail = getOnErrorFailAttribute(ifElement, defaultOnErrorFail);

      if (ifElement.getAttribute(CWConstants.LABEL_NAME_ATTRIBUTE) != null)
         io.labelName = ifElement.getAttribute(CWConstants.LABEL_NAME_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(CWConstants.ACTION_NAME_ATTRIBUTE) != null)
         io.ifActionName = ifElement.getAttribute(CWConstants.ACTION_NAME_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(CWConstants.FIELD_NAME_ATTRIBUTE) != null)
         io.ifFieldName = ifElement.getAttribute(CWConstants.FIELD_NAME_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(CWConstants.OPERATOR_ATTRIBUTE) != null)
         io.operator = ifElement.getAttribute(CWConstants.OPERATOR_ATTRIBUTE).getValue().trim();

      if (ifElement.getAttribute(CWConstants.VALUE_ATTRIBUTE) != null)
         io.value = ifElement.getAttribute(CWConstants.VALUE_ATTRIBUTE).getValue().trim();

      if ( (actionsElement = ifElement.getChild(CWConstants.ACTIONS_ELEMENT)) != null)
         io.actionAL = extractActions(actionsElement);

      if ( (elseElement = ifElement.getChild(CWConstants.ELSE_ELEMENT)) != null)
         io.eo = (ElseObject)retrieveElse(elseElement);

      return io;
   }

   /**
    * Retrieve all XML values for else action.
    * @param elseElement
    * @throws JDOMException
    * @throws IOException 
    */
   private UFMObject retrieveElse(Element elseElement) throws JDOMException, IOException
   {
      Element actionsElement;
      ElseObject eo = new ElseObject();

      eo.onErrorFail = getOnErrorFailAttribute(elseElement, defaultOnErrorFail);

      if (elseElement.getAttribute(CWConstants.EXIT_RC_ATTRIBUTE) != null)
         eo.exitRC = StringUtilities.stringToInt(elseElement.getAttribute(CWConstants.EXIT_RC_ATTRIBUTE).getValue().trim(), 0);

      if ( (actionsElement = elseElement.getChild(CWConstants.ACTIONS_ELEMENT)) != null)
         eo.actionAL = extractActions(actionsElement);

      return eo;
   }

   /**
    * Retrieve all XML values for Loop action.
    * @param actionElement
    * @throws JDOMException
    * @throws IOException 
    */
   private UFMObject retrieveLoop(Element actionElement) throws JDOMException, IOException
   {
      Element actionsElement;
      LoopObject lo = new LoopObject();
      String num = "";

      lo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.START_ATTRIBUTE) != null)
      {
         num = actionElement.getAttribute(CWConstants.START_ATTRIBUTE).getValue().trim();
         lo.start = StringUtilities.stringToInt(num, 1);
      }

      if (actionElement.getAttribute(CWConstants.END_ATTRIBUTE) != null)
      {
         num = actionElement.getAttribute(CWConstants.END_ATTRIBUTE).getValue().trim();
         lo.end = StringUtilities.stringToInt(num, 1);
      }

      if (actionElement.getAttribute(CWConstants.BY_ATTRIBUTE) != null)
      {
         num = actionElement.getAttribute(CWConstants.BY_ATTRIBUTE).getValue().trim();
         lo.by = StringUtilities.stringToInt(num, 1);
      }

      if ( (actionsElement = actionElement.getChild(CWConstants.ACTIONS_ELEMENT)) != null)
         lo.actionAL = extractActions(actionsElement);

      return lo;
   }

   /**
    * Retrieve all XML values for Launch action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveLaunch(Element actionElement)
   {
      LaunchObject so = new LaunchObject();

      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
         so.xmlFile = actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();

      return so;
   }

   /**
    * Retrieve all XML values for Schedule action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveSchedule(Element actionElement)
   {
      ScheduleObject so = new ScheduleObject();

      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
         so.xmlFile = actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();

      return so;
   }

   /**
    * Retrieve all XML values for SendEmail action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveSendEmail(Element actionElement)
   {
      SendEmailObject seo = new SendEmailObject();
      
      seo.labelName = getActionLabelAttribute(actionElement);

      seo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
         seo.xmlFile = actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();

      return seo;
   }

   /**
    * Retrieve all XML values for sleep action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveSleep(Element actionElement)
   {
      SleepObject so = new SleepObject();
      
      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.HOURS_ATTRIBUTE) != null)
         so.hours = StringUtilities.stringToInt(actionElement.getAttribute(CWConstants.HOURS_ATTRIBUTE).getValue().trim(), 0);

      if (actionElement.getAttribute(CWConstants.MINUTES_ATTRIBUTE) != null)
         so.minutes = StringUtilities.stringToInt(actionElement.getAttribute(CWConstants.MINUTES_ATTRIBUTE).getValue().trim(), 0);

      if (actionElement.getAttribute(CWConstants.SECONDS_ATTRIBUTE) != null)
         so.seconds = StringUtilities.stringToInt(actionElement.getAttribute(CWConstants.SECONDS_ATTRIBUTE).getValue().trim(), 0);

      if (actionElement.getAttribute(CWConstants.MILLISECONDS_ATTRIBUTE) != null)
         so.milliseconds = StringUtilities.stringToInt(actionElement.getAttribute(CWConstants.MILLISECONDS_ATTRIBUTE).getValue().trim(), 0);

      return so;
   }

   /**
    * Retrieve all XML values for Append action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveAppend(Element actionElement)
   {
      Element fileElement = null;
      
      AppendObject ao = new AppendObject();
      
      ao.labelName = getActionLabelAttribute(actionElement);

      ao.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            ao.createDir = true;
      }

      ao.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         ao.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (ao.toFile != null) && (ao.toFile.indexOf(File.separator) != -1) )
      {
         int xx = ao.toFile.lastIndexOf(File.separator);
         ao.toDir  = ao.toFile.substring(0, xx);
         ao.toFile = ao.toFile.substring(xx+1);
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            ao.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            ao.srcFilename = filename;
      }

      return ao;
   }

   /**
    * Retrieve all XML values for convert action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveConvert(Element actionElement)
   {
      Element fileElement = null;

      ConvertObject co = new ConvertObject();

      co.labelName = getActionLabelAttribute(actionElement);

      co.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      co.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE) != null)
         co.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if ( (co.toFile != null) && (co.toFile.indexOf(File.separator) != -1) )
      {
         int xx = co.toFile.lastIndexOf(File.separator);
         co.toDir  = co.toFile.substring(0, xx);
         co.toFile = co.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            co.createDir = true;
      }

      if (actionElement.getAttribute(CWConstants.MODE_ATTRIBUTE) != null)
      {
         String mode = actionElement.getAttribute(CWConstants.MODE_ATTRIBUTE).getValue().trim();
         
         for (int i=0; i < CWConstants.CONVERT_ARRAY.length; i++)
         {
            if (CWConstants.CONVERT_ARRAY[i].equalsIgnoreCase(mode))
            {
               co.mode = i;
               break;
            }
         }
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            co.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            co.srcFilename = filename;
      }

      return co;
   }

   /**
    * Retrieve all XML values for copy action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveCopy(Element actionElement)
   {
      Element fileElement = null;
      
      CopyObject co = new CopyObject();
      
      co.labelName = getActionLabelAttribute(actionElement);

      co.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      co.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE) != null)
         co.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if ( (co.toFile != null) && (co.toFile.indexOf(File.separator) != -1) )
      {
         int xx = co.toFile.lastIndexOf(File.separator);
         co.toDir  = co.toFile.substring(0, xx);
         co.toFile = co.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            co.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            co.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            co.srcFilename = filename;
      }

      return co;
   }

   /**
    * Retrieve all XML values for DecryptFile action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveDecryptFile(Element actionElement)
   {
      Element fileElement = null;

      DecryptFileObject dfo = new DecryptFileObject();
      
      dfo.labelName = getActionLabelAttribute(actionElement);

      dfo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      dfo.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         dfo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (dfo.toFile != null) && (dfo.toFile.indexOf(File.separator) != -1) )
      {
         int xx = dfo.toFile.lastIndexOf(File.separator);
         dfo.toDir  = dfo.toFile.substring(0, xx);
         dfo.toFile = dfo.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE) != null)
          dfo.crypto.passPhrase = actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE) != null)
      {
          String keySize = actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE).getValue().trim();
          dfo.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            dfo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            dfo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            dfo.srcFilename = filename;
      }

      return dfo;
   }

   /**
    * Retrieve all XML values for delete action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveDelete(Element actionElement)
   {
      DeleteObject dlo = new DeleteObject();
      
      dlo.labelName = getActionLabelAttribute(actionElement);

      dlo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      dlo.filenames = getXMLFileTags(actionElement);
      
      return dlo;
   }

   /**
    * Retrieve all XML values for EncryptFile action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveEncryptFile(Element actionElement)
   {
      Element fileElement = null;

      EncryptFileObject efo = new EncryptFileObject();
      
      efo.labelName = getActionLabelAttribute(actionElement);

      efo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         efo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      efo.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE) != null)
      {
          efo.crypto.passPhrase = actionElement.getAttribute(CWConstants.PASS_PHRASE_ATTRIBUTE).getValue().trim();
      }

      if (actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE) != null)
      {
          String keySize = actionElement.getAttribute(CWConstants.KEY_SIZE_ATTRIBUTE).getValue().trim();
          efo.crypto.keySize = StringUtilities.stringToInt(keySize, 128);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            efo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            efo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            efo.srcFilename = filename;
      }

      return efo;
   }

   /**
    * Retrieve all XML values for MakeDir action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveMakeDir(Element actionElement)
   {
      MakeDirObject mdo = new MakeDirObject();

      mdo.labelName = getActionLabelAttribute(actionElement);

      mdo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      mdo.targetDir = actionElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim();

      return mdo;
   }

   /**
    * Retrieve all XML values for Merge action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveMerge(Element actionElement)
   {
      MergeObject mo = new MergeObject();
      
      mo.labelName = getActionLabelAttribute(actionElement);

      mo.srcFilenames = getXMLFileTags(actionElement);

      mo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      mo.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         mo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (mo.toFile != null) && (mo.toFile.indexOf(File.separator) != -1) )
      {
         int xx = mo.toFile.lastIndexOf(File.separator);
         mo.toDir  = mo.toFile.substring(0, xx);
         mo.toFile = mo.toFile.substring(xx+1);
      }
      
      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            mo.createDir = true;
      }
      
      return mo;
   }

   /**
    * Retrieve all XML values for MergeSort action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveMergeSort(Element actionElement)
   {
      MergeSortObject mso = new MergeSortObject();
      
      mso.labelName = getActionLabelAttribute(actionElement);

      mso.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.COLUMN_ATTRIBUTE) != null)
      {
          String column = actionElement.getAttribute(CWConstants.COLUMN_ATTRIBUTE).getValue().trim();
          mso.column = StringUtilities.stringToInt(column, 1);
      }
      
      if (actionElement.getAttribute(CWConstants.ORDER_ATTRIBUTE) != null)
          mso.order = actionElement.getAttribute(CWConstants.ORDER_ATTRIBUTE).getValue().trim().charAt(0);

      mso.srcFilenames = getXMLFileTags(actionElement);

      mso.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         mso.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         mso.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (mso.toFile != null) && (mso.toFile.indexOf(File.separator) != -1) )
      {
         int xx = mso.toFile.lastIndexOf(File.separator);
         mso.toDir  = mso.toFile.substring(0, xx);
         mso.toFile = mso.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            mso.createDir = true;
      }

      return mso;
   }

   /**
    * Retrieve all XML values for move action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveMove(Element actionElement)
   {
      Element fileElement = null;
      
      MoveObject mo = new MoveObject();
      
      mo.labelName = getActionLabelAttribute(actionElement);

      mo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if ((fileElement =  actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            mo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            mo.srcFilename = filename;
      }

      mo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE) != null)
         mo.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if ( (mo.toFile != null) && (mo.toFile.indexOf(File.separator) != -1) )
      {
         int xx = mo.toFile.lastIndexOf(File.separator);
         mo.toDir  = mo.toFile.substring(0, xx);
         mo.toFile = mo.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            mo.createDir = true;
      }

      return mo;
   }

   /**
    * Retrieve all XML values for Re-Encode action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveReEncode(Element actionElement)
   {
      Element fileElement = null;

      ReEncodeObject reo = new ReEncodeObject();

      reo.labelName = getActionLabelAttribute(actionElement);

      reo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      reo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE) != null)
         reo.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if ( (reo.toFile != null) && (reo.toFile.indexOf(File.separator) != -1) )
      {
         int xx = reo.toFile.lastIndexOf(File.separator);
         reo.toDir  = reo.toFile.substring(0, xx);
         reo.toFile = reo.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            reo.createDir = true;
      }

      if (actionElement.getAttribute(CWConstants.INPUT_ENCODING_ATTRIBUTE) != null)
         reo.inputEncoding = actionElement.getAttribute(CWConstants.INPUT_ENCODING_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.OUTPUT_ENCODING_ATTRIBUTE) != null)
         reo.outputEncoding = actionElement.getAttribute(CWConstants.OUTPUT_ENCODING_ATTRIBUTE).getValue().trim();

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            reo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            reo.srcFilename = filename;
      }

      return reo;
   }

   /**
    * Retrieve all XML values for remove items action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveRemoveItems(Element actionElement)
   {
      Element fileElement = null;

      RemoveItemsObject rio = new RemoveItemsObject();

      rio.labelName = getActionLabelAttribute(actionElement);

      rio.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      rio.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE) != null)
         rio.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if ( (rio.toFile != null) && (rio.toFile.indexOf(File.separator) != -1) )
      {
         int xx = rio.toFile.lastIndexOf(File.separator);
         rio.toDir  = rio.toFile.substring(0, xx);
         rio.toFile = rio.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            rio.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            rio.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            rio.srcFilename = filename;
      }

      if ((fileElement = actionElement.getChild(CWConstants.ITEM_FILE_ELEMENT)) != null)
      {
         String itemFilename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            rio.itemFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), itemFilename);
         else
            rio.itemFilename = itemFilename;
      }

      return rio;
   }

   /**
    * Retrieve all XML values for RemoveDir action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveRemoveDir(Element actionElement)
   {
      char o = ' ';
      RemoveDirObject rdo = new RemoveDirObject();

      rdo.labelName = getActionLabelAttribute(actionElement);

      rdo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      rdo.targetDir = actionElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.REMOVE_ALL_ATTRIBUTE) != null)
          o = actionElement.getAttribute(CWConstants.REMOVE_ALL_ATTRIBUTE).getValue().trim().charAt(0);

      if (o ==  ' ')
         rdo.removeAll = false;
      else if (o == 'Y')
         rdo.removeAll = true;
      else if (o == 'N')
         rdo.removeAll = false;

      return rdo;
   }

   /**
    * Retrieve all XML values for rename action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveRename(Element actionElement)
   {
      Element fileElement = null;
      
      RenameObject ro = new RenameObject();
      
      ro.labelName = getActionLabelAttribute(actionElement);

      ro.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      ro.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            ro.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            ro.srcFilename = filename;
      }

      return ro;
   }

   /**
    * Retrieve all XML values for ReplaceText action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveReplaceText(Element actionElement)
   {
      Element fileElement = null;
      Element findElement = null;
      Element replaceElement = null;

      ReplaceTextObject rto = new ReplaceTextObject();
      
      rto.labelName = getActionLabelAttribute(actionElement);

      rto.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      rto.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         rto.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (rto.toFile != null) && (rto.toFile.indexOf(File.separator) != -1) )
      {
         int xx = rto.toFile.lastIndexOf(File.separator);
         rto.toDir  = rto.toFile.substring(0, xx);
         rto.toFile = rto.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.ROWS_ATTRIBUTE) != null)
      {
         String rows = actionElement.getAttribute(CWConstants.ROWS_ATTRIBUTE).getValue().trim();
         rto.convertRTValue(rows, rto.row);
      }

      if (actionElement.getAttribute(CWConstants.COLUMNS_ATTRIBUTE) != null)
      {
          String columns = actionElement.getAttribute(CWConstants.COLUMNS_ATTRIBUTE).getValue().trim();
          rto.convertRTValue(columns, rto.column);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            rto.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            rto.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            rto.srcFilename = filename;
      }
      
      if ((findElement = actionElement.getChild(CWConstants.FIND_ELEMENT)) != null)
         rto.findText = findElement.getText();
      
      if ((replaceElement = actionElement.getChild(CWConstants.REPLACE_ELEMENT)) != null)
         rto.replaceText = replaceElement.getText();

      return rto;
   }

   /**
    * Retrieve all XML values for Scp action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveScp(Element actionElement)
   {
      ScpObject fo = new ScpObject();

      fo.labelName = getActionLabelAttribute(actionElement);

      fo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE) != null)
         fo.xmlFile = actionElement.getAttribute(CWConstants.XML_FILE_ATTRIBUTE).getValue().trim();

      return fo;
   }

   /**
    * Retrieve all XML values for Sort action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveSort(Element actionElement)
   {
      Element fileElement = null;
      
      SortObject so = new SortObject();
      
      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      so.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         so.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (so.toFile != null) && (so.toFile.indexOf(File.separator) != -1) )
      {
         int xx = so.toFile.lastIndexOf(File.separator);
         so.toDir  = so.toFile.substring(0, xx);
         so.toFile = so.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.COLUMN_ATTRIBUTE) != null)
      {
          String column = actionElement.getAttribute(CWConstants.COLUMN_ATTRIBUTE).getValue().trim();
          so.column = StringUtilities.stringToInt(column, 1);
      }
      
      if (actionElement.getAttribute(CWConstants.ORDER_ATTRIBUTE) != null)
          so.order = actionElement.getAttribute(CWConstants.ORDER_ATTRIBUTE).getValue().trim().toUpperCase().charAt(0);

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            so.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            so.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            so.srcFilename = filename;
      }

      return so;
   }

   /**
    * Retrieve all XML values for SortUnique action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveSortUnique(Element actionElement)
   {
      Element fileElement = null;

      SortUniqueObject suo = new SortUniqueObject();

      suo.labelName = getActionLabelAttribute(actionElement);

      suo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      suo.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         suo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (suo.toFile != null) && (suo.toFile.indexOf(File.separator) != -1) )
      {
         int xx = suo.toFile.lastIndexOf(File.separator);
         suo.toDir  = suo.toFile.substring(0, xx);
         suo.toFile = suo.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            suo.createDir = true;
      }

      if (actionElement.getAttribute(CWConstants.COLUMN_ATTRIBUTE) != null)
      {
          String column = actionElement.getAttribute(CWConstants.COLUMN_ATTRIBUTE).getValue().trim();
          suo.column = StringUtilities.stringToInt(column, 1);
      }

      if (actionElement.getAttribute(CWConstants.ORDER_ATTRIBUTE) != null)
          suo.order = actionElement.getAttribute(CWConstants.ORDER_ATTRIBUTE).getValue().trim().toUpperCase().charAt(0);

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            suo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            suo.srcFilename = filename;
      }

      return suo;
   }

   /**
    * Retrieve all XML values for Split action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveSplit(Element actionElement)
   {
      Element fileElement = null;

      SplitObject so = new SplitObject();

      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         so.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      
      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            so.createDir = true;
      }

      if (actionElement.getAttribute(CWConstants.TYPE_ATTRIBUTE) != null)
      {
          so.type = actionElement.getAttribute(CWConstants.TYPE_ATTRIBUTE).getValue().trim().toUpperCase().charAt(0);
      }

      if (actionElement.getAttribute(CWConstants.SIZE_ATTRIBUTE) != null)
      {
          String size = actionElement.getAttribute(CWConstants.SIZE_ATTRIBUTE).getValue().trim();
          so.size = StringUtilities.stringToInt(size, 1);
      }

      if (actionElement.getAttribute(CWConstants.PREFIX_ATTRIBUTE) != null)
      {
          so.prefix = actionElement.getAttribute(CWConstants.PREFIX_ATTRIBUTE).getValue().trim();
      }

      if (actionElement.getAttribute(CWConstants.EXTENSION_ATTRIBUTE) != null)
      {
          so.extension = actionElement.getAttribute(CWConstants.EXTENSION_ATTRIBUTE).getValue().trim();
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            so.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            so.srcFilename = filename;
      }

      return so;
   }

   /**
    * Retrieve all XML values for Tar action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveTar(Element actionElement)
   {
      Element fileElement = null;
      
      TarObject to = new TarObject();
      
      to.labelName = getActionLabelAttribute(actionElement);

      to.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      to.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         to.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (to.toFile != null) && (to.toFile.indexOf(File.separator) != -1) )
      {
         int xx = to.toFile.lastIndexOf(File.separator);
         to.toDir  = to.toFile.substring(0, xx);
         to.toFile = to.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            to.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            to.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            to.srcFilename = filename;
      }

      return to;
   }

   /**
    * Retrieve all XML values for touch action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveTouch(Element actionElement)
   {
      TouchObject to = new TouchObject();
      
      to.labelName = getActionLabelAttribute(actionElement);

      to.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            to.createDir = true;
      }

      to.filenames = getXMLFileTags(actionElement);

      return to;
   }

   /**
    * Retrieve all XML values for Unique action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveUnique(Element actionElement)
   {
      Element fileElement = null;

      UniqueObject so = new UniqueObject();

      so.labelName = getActionLabelAttribute(actionElement);

      so.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      so.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         so.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if ( (so.toFile != null) && (so.toFile.indexOf(File.separator) != -1) )
      {
         int xx = so.toFile.lastIndexOf(File.separator);
         so.toDir  = so.toFile.substring(0, xx);
         so.toFile = so.toFile.substring(xx+1);
      }

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            so.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            so.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            so.srcFilename = filename;
      }

      return so;
   }

   /**
    * Retrieve all XML values for UnTar action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveUnTar(Element actionElement)
   {
      Element fileElement = null;
      
      UnTarObject uto = new UnTarObject();
      
      uto.labelName = getActionLabelAttribute(actionElement);

      uto.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      uto.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            uto.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String tarFilename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            uto.tarFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), tarFilename);
         else
            uto.tarFilename = tarFilename;
      }

      return uto;
   }

   /**
    * Retrieve all XML values for UnZip action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveUnZip(Element actionElement)
   {
      Element fileElement = null;

      UnZipObject uzo = new UnZipObject();
      
      uzo.labelName = getActionLabelAttribute(actionElement);

      uzo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      uzo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            uzo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String zipFilename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            uzo.zipFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), zipFilename);
         else
            uzo.zipFilename = zipFilename;
      }

      return uzo;
   }

   /**
    * Retrieve all XML values for watch action.
    * @param actionElement
    * @throws JDOMException
    * @throws IOException 
    */
   private UFMObject retrieveWatch(Element actionElement) throws JDOMException, IOException
   {
      Element actionsElement;
      WatchObject wo = new WatchObject();

      wo.labelName = getActionLabelAttribute(actionElement);

      wo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.POLL_INTERVAL_ATTRIBUTE) != null)
      {
         String pollInt = actionElement.getAttribute(CWConstants.POLL_INTERVAL_ATTRIBUTE).getValue().trim();
         wo.pollInterval = StringUtilities.stringToInt(pollInt, 15) * 1000; /* milli-seconds */
      }

      // read and parse watches
      List watchList = actionElement.getChildren(CWConstants.WATCH_ITEM_ELEMENT);
      wo.watches = new Watch[watchList.size()];
      int wi = 0;
      Iterator it = watchList.iterator();
      while (it.hasNext())
      {
         Element watchElement = (Element) it.next();
         String type = watchElement.getAttribute(CWConstants.TYPE_ATTRIBUTE).getValue();

         // create watch objects
         Watch t_watch = null;
         // file type of watch: create FileWatch instance
         if ("F".equalsIgnoreCase(type))
         {
            Element fileElement = watchElement.getChild(CWConstants.FILE_ELEMENT);
            String filename = fileElement.getTextTrim();

            String watchfilename;

            if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
               watchfilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue(), filename);
            else
               watchfilename = filename;

            watchfilename = Tokens.replace(watchfilename, _properties);

            t_watch = new FileWatch(watchfilename);
         }
         // directory watch type: parse all extensions and create DirectoryWatch instance
         else if ("D".equalsIgnoreCase(type))
         {
            Element directoryElement = watchElement.getChild(CWConstants.DIRECTORY_ELEMENT);
            String watchDirectory = directoryElement.getTextTrim();

            boolean sendFlag = false;
            if (directoryElement.getAttribute(CWConstants.SEND_ON_STARTUP).getValue().toUpperCase().charAt(0) == 'Y')
               sendFlag = true;
            else
               sendFlag = false;

            // check if file exists:
            File file = new File(watchDirectory);
            if (!file.exists())
               throw new IOException("Watch: " + watchDirectory + " not found!");

            List extensionList = watchElement.getChildren(CWConstants.EXTENSION_ELEMENT);
            String[] filter;
            /* Did the user specify any extensions? */
            if (extensionList.size() > 0)
            {
               filter = new String[extensionList.size()];
               Iterator it2 = extensionList.iterator();
               int f = 0;
               while (it2.hasNext())
               {
                  Element extensionElement = (Element) it2.next();
                  filter[f++] = extensionElement.getTextTrim();
               }
            }
            else
            {  /* set a default if the user forget to specify an extension */
               filter = new String[1];
               filter[0] = "*";
            }

            watchDirectory = Tokens.replace(watchDirectory, _properties);

            t_watch = new DirectoryWatch(watchDirectory, filter, sendFlag);
         }

         if ( (actionsElement = watchElement.getChild(CWConstants.ACTIONS_ELEMENT)) != null)
            t_watch.actionAL = extractActions(actionsElement);
         
         wo.watches[wi++] = t_watch;
      }

      return wo;
   }

   /**
    * Retrieve all XML values for WriteText action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveWriteText(Element actionElement)
   {
      char o = ' ';
      WriteTextObject wto = new WriteTextObject();

      wto.labelName = getActionLabelAttribute(actionElement);

      wto.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         wto.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE) != null)
         wto.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.APPEND_ATTRIBUTE) != null)
          o = actionElement.getAttribute(CWConstants.APPEND_ATTRIBUTE).getValue().trim().toUpperCase().charAt(0);

      if (o ==  ' ')
         wto.append = false;
      else if (o == 'Y')
         wto.append = true;
      else if (o == 'N')
         wto.append = false;

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            wto.createDir = true;
      }

      wto.text = actionElement.getTextTrim();

      return wto;
   }

   /**
    * Retrieve all XML values for Zip action.
    * @param actionElement
    * @throws JDOMException
    */
   private UFMObject retrieveZip(Element actionElement)
   {
      Element fileElement = null;

      ZipObject zo = new ZipObject();
      
      zo.labelName = getActionLabelAttribute(actionElement);

      zo.onErrorFail = getOnErrorFailAttribute(actionElement, defaultOnErrorFail);

      if (actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE) != null)
         zo.toDir = actionElement.getAttribute(CWConstants.TO_DIR_ATTRIBUTE).getValue().trim();

      zo.toFile = actionElement.getAttribute(CWConstants.TO_FILE_ATTRIBUTE).getValue().trim();

      if (actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE) != null)
      {
         String flag = actionElement.getAttribute(CWConstants.CREATE_DIR_ATTRIBUTE).getValue().toUpperCase();
         if (flag.charAt(0) == CWConstants.YES_FLAG)
            zo.createDir = true;
      }

      if ((fileElement = actionElement.getChild(CWConstants.FILE_ELEMENT)) != null)
      {
         String filename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            zo.srcFilename = IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), filename);
         else
            zo.srcFilename = filename;
      }

      return zo;
   }

   /**
    * Retrieve all XML values for MQ tag.
    * @param actionElement
    * @throws JDOMException
    */
   private Element retrieveMQ(Element actionElement) throws JDOMException
   {
      Element mqElement = null;
      Element queueElement = null;
      
      mqXMLFile= defaultMQXmlFile;
      CCDTFile = defaultCCDTFile;
      qMgrName = defaultQMgrName;
      qObject  = defaultQObject;

      if ((mqElement = actionElement.getChild(CWConstants.MQ_ELEMENT)) != null)
      {
         if ( mqElement.getChild(CWConstants.MQFILE_ELEMENT) != null)
         {
            mqXMLFile = mqElement.getChild(CWConstants.MQFILE_ELEMENT).getTextTrim();

            boolean exists = (new File(CWConstants.MQ_PATH + mqXMLFile)).exists();
            if ( !(exists) )
               throw new JDOMException("Error: The MQ XML file does not exists : " + CWConstants.MQ_PATH + mqXMLFile);
         }

         if ( mqElement.getChild(CWConstants.CCDTFILE_ELEMENT) != null)
            CCDTFile = mqElement.getChild(CWConstants.CCDTFILE_ELEMENT).getTextTrim();

         if ( (mqXMLFile != null) && (CCDTFile != null) )
            throw new JDOMException("Error: Do not specify both "+CWConstants.MQFILE_ELEMENT+" and "+CWConstants.CCDTFILE_ELEMENT+" elements.");

         if ( mqElement.getChild(CWConstants.QMGRNAME_ELEMENT) != null)
            qMgrName = mqElement.getChild(CWConstants.QMGRNAME_ELEMENT).getTextTrim();
         
         if ( (queueElement = mqElement.getChild(CWConstants.QUEUENAME_ELEMENT)) != null)
         {
            qObject = new QueueObject();

            qObject.qName = queueElement.getTextTrim();

            if (queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE) != null)
               qObject.qMgrName = queueElement.getAttribute(CWConstants.QMGRNAME_ATTRIBUTE).getValue().trim();
            else
               qObject.qMgrName = null;
         }
      }

      if (mqXMLFile != null)
         mqXMLFile  = Tokens.replace(mqXMLFile, _properties);
      if (CCDTFile != null)
         CCDTFile   = Tokens.replace(CCDTFile, _properties);
      if (qMgrName != null)
         qMgrName   = Tokens.replace(qMgrName, _properties);
      if (qObject != null)
      {
         if (qObject.qName != null)
            qObject.qName      = Tokens.replace(qObject.qName, _properties);
         if (qObject.qMgrName != null)
            qObject.qMgrName      = Tokens.replace(qObject.qMgrName, _properties);
      }

      return mqElement;
   }

   /**
    * Retrieve 1 or more XML File elements from the section.
    * @param actionElement
    * @return newFilenames
    */
   private String[] getXMLFileTags(Element actionElement)
   {
      Element fileElement = null;
      ArrayList<String> al = new ArrayList<String>();
      String srcFilename = null;

      /* We should have at least 2 File elements */
      List propertyList = actionElement.getChildren(CWConstants.FILE_ELEMENT);
      Iterator it = propertyList.iterator();
      while (it.hasNext())
      {
         fileElement = (Element) it.next();

         srcFilename = fileElement.getTextTrim();

         if (fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE) != null)
            al.add(IO.apendFileToPath(fileElement.getAttribute(CWConstants.DIR_ATTRIBUTE).getValue().trim(), srcFilename));
         else
            al.add(srcFilename);
      }
      
      String[] newFilenames = new String[al.size()];
      for (int i=0; i < al.size(); i++)
         newFilenames[i] = al.get(i);
      
      return newFilenames;
   }

   /**
    * Retrieve
    * @param actionElement
    * @return newFilenames
    */
   private String getActionLabelAttribute(Element actionElement)
   {
      String name = "default";

      if (actionElement.getAttribute(CWConstants.LABEL_ATTRIBUTE) != null)
         name = actionElement.getAttribute(CWConstants.LABEL_ATTRIBUTE).getValue().trim();

      return name;
   }
   
   /**
    * Add server related tokens and their values to the hashtable.
    */
   private void addServerSettings()
   {
      try
      {
         String value = InetAddress.getLocalHost().getHostName();
         _properties.put(CWConstants.HOSTNAME_TOKEN, value);
      }
      catch (UnknownHostException e)
      {}

      try
      {
         String value = InetAddress.getLocalHost().getHostAddress();
         _properties.put(CWConstants.IPADDRESS_TOKEN, value);
      }
      catch (UnknownHostException e)
      {}
   }

   /**
    * Retrieve an ArrayList of all of the actions to be performed.
    * @return actionAL
    */
   public ArrayList<UFMObject> getActions()
   {
      return _actionAL;
   }

   /**
    * Retrieve an OnErrorObject.
    * @return onError
    */
   public OnErrorObject getOnError()
   {
      return _onError;
   }

   /**
    * Retrieve an statusObject.
    * @return statusObject
    */
   public StatusObject getStatusObject()
   {
      return statusObject;
   }

   /**
    * Retrieve user properties.
    * @return properties
    */
   public Hashtable<String,String>  getProperties()
   {
      return _properties;
   }
}
